"""Utilities for querying and matching in the PostgreSQL backend."""

from typing import Literal, TypeVar

import pyarrow as pa
from sqlalchemy import (
    CTE,
    ColumnElement,
    FromClause,
    and_,
    func,
    join,
    outerjoin,
    select,
    union_all,
)
from sqlalchemy.orm import Session, aliased
from sqlalchemy.sql.selectable import Select

from matchbox.common.db import sql_to_df
from matchbox.common.dtos import Match
from matchbox.common.exceptions import (
    MatchboxResolutionNotFoundError,
)
from matchbox.common.graph import ResolutionName, SourceResolutionName
from matchbox.common.logging import logger
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    Probabilities,
    ResolutionFrom,
    Resolutions,
    SourceConfigs,
)
from matchbox.server.postgresql.utils.db import compile_sql

T = TypeVar("T")


def get_source_config(name: SourceResolutionName, session: Session) -> SourceConfigs:
    """Converts the named source to a SourceConfigs ORM object."""
    source_config = (
        session.query(SourceConfigs)
        .join(Resolutions, Resolutions.resolution_id == SourceConfigs.resolution_id)
        .filter(Resolutions.name == name)
        .first()
    )
    if source_config is None:
        raise MatchboxResolutionNotFoundError(name=name)

    return source_config


def _build_probability_subquery(
    res_id: int, threshold_val: float, subquery_name: str
) -> Select:
    """Build a probability subquery for a given resolution.

    When a leaf belongs to multiple clusters that meet the threshold,
    this selects the cluster with probability CLOSEST to the threshold
    (i.e., the most conservative/weakest clustering decision).
    """
    # Create unique aliases for this subquery to avoid parameter conflicts
    contains_alias = aliased(Contains, name=f"contains_{subquery_name}")
    probabilities_alias = aliased(Probabilities, name=f"prob_{subquery_name}")

    return (
        select(
            contains_alias.leaf.label("leaf"),
            contains_alias.root.label(f"cluster_{subquery_name}"),
            probabilities_alias.probability,
        )
        .select_from(
            join(
                contains_alias,
                probabilities_alias,
                contains_alias.root == probabilities_alias.cluster_id,
            )
        )
        .where(
            and_(
                probabilities_alias.resolution_id == res_id,
                probabilities_alias.probability >= threshold_val,
            )
        )
        .distinct(contains_alias.leaf)
        .order_by(
            contains_alias.leaf,
            probabilities_alias.probability.asc(),
            contains_alias.root.asc(),
        )
        .subquery(f"prob_{subquery_name}")
    )


def build_unified_query(
    resolution: Resolutions,
    sources: list[SourceConfigs] | None = None,
    threshold: int | None = None,
    level: Literal["leaf", "key"] = "leaf",
    get_hashes: bool = False,
) -> Select:
    """Build a query to resolve cluster assignments across resolution hierarchies.

    This function creates SQL that determines which cluster each source record belongs
    to by traversing up a resolution hierarchy and applying priority-based cluster
    selection.

    The query uses `COALESCE` to implement a priority system where higher-level
    resolutions can "claim" records, with lower levels only processing unclaimed
    records:

    ```sql
    COALESCE(highest_priority_cluster, medium_priority_cluster, ..., source_cluster)
    ```

    1. **Lineage discovery**: Queries the resolution hierarchy to find all ancestor
        resolutions, ordered by priority (lowest level = highest priority)
    2. **Source filtering**: When `sources` is provided, constrains results to only
        include clusters from those specific source configurations
    3. **Threshold application**: Applies probability thresholds to determine which
        clusters qualify at each resolution level
    4. **Subquery construction**: For each model resolution in the lineage, builds
        a subquery that finds qualifying clusters via the Contains→Probabilities
        join. Each joined subquery adds a new cluster column which is then merged
        via...
    5. **`COALESCE` assembly**: Joins all subqueries to source data and uses `COALESCE`
        to select the highest-priority cluster assignment for each record

    The level changes the data returned:

    * `"leaf"`: Returns both root and leaf cluster IDs. For unmerged source
        clusters, the root and leaf properties will be the same.
    * `"key"`: In addition to the above, it also returns the source key. This will give
        more rows than `"leaf"` because it needs a row for every key attached to a leaf.

    Additionally, if `get_hashes` is set to True, the root and leaf hashes are returned.
    """
    # Get ordered lineage (already sorted by priority)
    lineage = resolution.get_lineage(sources=sources, threshold=threshold)

    # Separate to model and source resolutions
    model_resolutions: list[tuple[int, float]] = []
    source_config_ids: list[int] = []
    for resolution_id, source_config_id, truth in lineage:
        if truth is None:
            source_config_ids.append(source_config_id)
        else:
            model_resolutions.append((resolution_id, truth))

    # Build source config filter
    if sources:
        # If sources are provided, filter to those source configs
        source_config_ids = set(sc.source_config_id for sc in sources) & set(
            source_config_ids
        )
        source_filter = ClusterSourceKey.source_config_id.in_(source_config_ids)
    elif resolution.type == "source":
        # If querying a source resolution with no sources filter,
        # filter to just that source
        source_filter = (
            ClusterSourceKey.source_config_id
            == resolution.source_config.source_config_id
        )
    else:
        # No sources provided, filter to lineage source configs
        source_filter = ClusterSourceKey.source_config_id.in_(source_config_ids)

    # Create proper aliases for clusters tables
    leaf_clusters = aliased(Clusters, name="leaf_clusters")
    root_clusters = aliased(Clusters, name="root_clusters")

    # `ClusterSourceKey` is the basis for all subsequent joins
    from_clause: FromClause = ClusterSourceKey

    # Handle source-only case (no model resolutions in lineage)
    if not model_resolutions:
        # We always must select from `ClusterSourceKey`, as it points to source clusters
        selection = [
            ClusterSourceKey.cluster_id.label("root_id"),
            ClusterSourceKey.cluster_id.label("leaf_id"),
        ]

        if level == "key":
            selection.append(
                ClusterSourceKey.key,
            )

        if get_hashes:
            selection += [
                leaf_clusters.cluster_hash.label("root_hash"),
                leaf_clusters.cluster_hash.label("leaf_hash"),
            ]

            from_clause = join(
                from_clause,
                leaf_clusters,
                ClusterSourceKey.cluster_id == leaf_clusters.cluster_id,
            )

    else:  # Querying from a resolution not at the bottom
        # Build subqueries for each model resolution
        # Note both subqueries and cluster_columns are in priority order
        subqueries: list[Select] = []
        cluster_columns: list[ColumnElement] = []

        for i, (res_id, threshold_val) in enumerate(model_resolutions):
            subquery = _build_probability_subquery(res_id, threshold_val, str(i))
            subqueries.append(subquery)
            cluster_columns.append(subquery.c[f"cluster_{i}"])

        # To get hashes we need to join `Clusters`, here for leaves
        # and later for roots
        if get_hashes:
            from_clause: FromClause = join(
                from_clause,
                leaf_clusters,
                ClusterSourceKey.cluster_id == leaf_clusters.cluster_id,
            )

        # Add LEFT JOINs for each model resolution subquery
        for subquery in subqueries:
            from_clause = outerjoin(
                from_clause,
                subquery,
                ClusterSourceKey.cluster_id == subquery.c.leaf,
            )

        # Build final cluster ID
        # First non-null wins (highest priority first)
        final_root_cluster_id = func.coalesce(
            *cluster_columns, ClusterSourceKey.cluster_id
        )

        selection = [
            final_root_cluster_id.label("root_id"),
            ClusterSourceKey.cluster_id.label("leaf_id"),
        ]

        if level == "key":
            selection.append(
                ClusterSourceKey.key,
            )

        if get_hashes:
            selection += [
                root_clusters.cluster_hash.label("root_hash"),
                leaf_clusters.cluster_hash.label("leaf_hash"),
            ]

            from_clause: FromClause = from_clause.outerjoin(
                root_clusters, final_root_cluster_id == root_clusters.cluster_id
            )

    query = select(*selection).select_from(from_clause).where(source_filter)

    # Because we start from `ClusterSourceKey`, we must remove duplicates caused
    # by distinct keys on the same leaf
    if level == "leaf":
        query = query.distinct()

    return query


def _build_target_cluster_cte(
    key: str,
    source_config_id: int,
    resolution: Resolutions,
    threshold: int | None,
) -> Select:
    """Build the target_cluster CTE.

    Follows very similar logic to `build_unified_query`, but with filtering
    specifically for a single key and source_config_id.
    """
    # Get ordered lineage
    lineage = resolution.get_lineage(threshold=threshold)
    model_resolutions = [
        (res_id, truth) for res_id, _, truth in lineage if truth is not None
    ]

    # Build subqueries for each model resolution
    # Note both subqueries and cluster_columns are in priority order
    subqueries = []
    cluster_columns = []
    for i, (res_id, threshold_val) in enumerate(model_resolutions):
        subquery = _build_probability_subquery(res_id, threshold_val, str(i))
        subqueries.append(subquery)
        cluster_columns.append(subquery.c[f"cluster_{i}"])

    # Build FROM clause starting with cluster_keys
    from_clause = ClusterSourceKey

    # Add LEFT JOINs for each resolution subquery
    for subquery in subqueries:
        from_clause = outerjoin(
            from_clause,
            subquery,
            ClusterSourceKey.cluster_id == subquery.c.leaf,
        )

    # Build final cluster ID using COALESCE - first non-null wins
    final_cluster_id = func.coalesce(*cluster_columns, ClusterSourceKey.cluster_id)

    return (
        select(final_cluster_id.label("cluster_id"))
        .select_from(from_clause)
        .where(
            and_(
                ClusterSourceKey.key == key,
                ClusterSourceKey.source_config_id == source_config_id,
            )
        )
    )


def _build_matching_leaves_cte(
    target_source_config_ids: list[int],
    resolution: Resolutions,
    threshold: int | None,
    target_cluster_cte: CTE,
) -> Select:
    """Find all keys from target sources that belong to the target cluster.

    Given a target cluster ID, find all keys from my target sources that belong to
    it through ANY path in the hierarchy.

    Uses `UNION ALL` to combine multiple ways a key can belong to the target cluster:

        1. **Direct membership**: Keys that directly belong to the target cluster ID
        2. **Hierarchy membership**: For each model resolution in the lineage, keys that
            are connected to the target cluster through Contains→Probabilities chains

    The target cluster ID comes from the target_cluster_cte, and we search for all
    keys from the specified target sources that are related to it through any path
    in the resolution hierarchy.

    Returns a union of all matching keys with their cluster_id, key, and
    source_config_id.
    """
    # Get ordered lineage and extract model resolutions
    lineage = resolution.get_lineage(threshold=threshold)
    model_resolutions = [
        (res_id, truth) for res_id, _, truth in lineage if truth is not None
    ]

    # Start with direct members branch
    branches = [
        select(
            ClusterSourceKey.cluster_id,
            ClusterSourceKey.key,
            ClusterSourceKey.source_config_id,
        )
        .select_from(
            ClusterSourceKey, target_cluster_cte
        )  # CROSS JOIN with target_cluster
        .where(
            and_(
                ClusterSourceKey.source_config_id.in_(target_source_config_ids),
                ClusterSourceKey.cluster_id == target_cluster_cte.c.cluster_id,
            )
        )
    ]

    # Add branches for each model resolution
    for res_id, threshold_val in model_resolutions:
        branch = (
            select(
                ClusterSourceKey.cluster_id,
                ClusterSourceKey.key,
                ClusterSourceKey.source_config_id,
            )
            .select_from(
                join(
                    join(
                        ClusterSourceKey,
                        Contains,
                        ClusterSourceKey.cluster_id == Contains.leaf,
                    ),
                    Probabilities,
                    and_(
                        Contains.root == Probabilities.cluster_id,
                        Probabilities.resolution_id == res_id,
                        Probabilities.probability >= threshold_val,
                    ),
                ),
                target_cluster_cte,  # CROSS JOIN with target_cluster
            )
            .where(
                and_(
                    ClusterSourceKey.source_config_id.in_(target_source_config_ids),
                    Probabilities.cluster_id == target_cluster_cte.c.cluster_id,
                )
            )
        )
        branches.append(branch)

    # Combine all branches with UNION ALL
    if len(branches) == 1:
        return branches[0]

    return union_all(*branches)


def _build_match_query(
    key: str,
    source_config_id: int,
    target_source_config_ids: list[int],
    resolution: Resolutions,
    threshold: int | None,
) -> Select:
    """Build a match query to find all keys that cluster with a given input key.

    This function creates SQL that identifies which cluster an input key belongs to,
    then finds all other keys from specified target sources that belong to the same
    cluster through the resolution hierarchy.

    The query uses two CTEs to solve the matching problem:

        1. **Target cluster identification**: Determines which cluster the input key
            belongs to at the specified resolution level
        2. **Matching leaves discovery**: Finds all keys from target sources that
            belong to the same target cluster

    The overall process:

        1. **Target cluster CTE**: Uses the same `COALESCE` hierarchy logic as
            `build_unified_query` to resolve which cluster the input key belongs to.
            This handles the full resolution hierarchy with proper priority ordering.
        2. **Matching leaves CTE**: Builds a `UNION ALL` query with multiple branches:

            - **Direct members**: Keys that belong directly to the target cluster
            - **Hierarchy branches**: For each model resolution, finds keys that are
                connected to the target cluster through the Contains→Probabilities joins

        3. **Final assembly**: `LEFT JOINs` the target cluster with matching leaves,
            ensuring we always get the cluster ID even if no target matches exist
    """
    target_cluster_cte = _build_target_cluster_cte(
        key=key,
        source_config_id=source_config_id,
        resolution=resolution,
        threshold=threshold,
    ).cte("target_cluster")

    # Include both source and target configs to get all keys in the cluster
    all_source_config_ids = [source_config_id] + target_source_config_ids
    matching_leaves_cte = _build_matching_leaves_cte(
        target_source_config_ids=all_source_config_ids,
        resolution=resolution,
        threshold=threshold,
        target_cluster_cte=target_cluster_cte,
    ).cte("matching_leaves")

    return (
        select(
            matching_leaves_cte.c.cluster_id.label("cluster"),
            matching_leaves_cte.c.source_config_id,
            matching_leaves_cte.c.key,
        )
        .select_from(matching_leaves_cte)
        .distinct()
    )


def query(
    source: SourceResolutionName,
    resolution: ResolutionName | None = None,
    threshold: int | None = None,
    return_leaf_id: bool = False,
    limit: int = None,
) -> pa.Table:
    """Queries Matchbox to retrieve linked data for a source.

    Retrieves all linked data for a given source, resolving through hierarchy if needed.

    * Simple case: If querying the same resolution as the source, just select cluster
        IDs and keys directly from ClusterSourceKey
    * Hierarchy case: Uses the unified query builder to traverse up the resolution
        hierarchy, applying COALESCE priority logic to determine which parent cluster
        each source record belongs to
    * Priority resolution: When multiple model resolutions could assign a record to
        different clusters, COALESCE ensures higher-priority resolutions win

    Returns all records with their final resolved cluster IDs.
    """
    with MBDB.get_session() as session:
        source_config: SourceConfigs = get_source_config(source, session)
        source_resolution: Resolutions = session.get(
            Resolutions, source_config.resolution_id
        )

        if resolution:
            truth_resolution: Resolutions = (
                session.query(Resolutions)
                .filter(Resolutions.name == resolution)
                .first()
            )
            if truth_resolution is None:
                raise MatchboxResolutionNotFoundError(name=resolution)
        else:
            truth_resolution: Resolutions = source_resolution

        id_query: Select = build_unified_query(
            resolution=truth_resolution,
            sources=[source_config],
            threshold=threshold,
            level="key",
        )

        if limit:
            id_query = id_query.limit(limit)

        with MBDB.get_adbc_connection() as conn:
            stmt: str = compile_sql(id_query)
            logger.debug(f"Query SQL: \n {stmt}")
            id_results = sql_to_df(
                stmt=stmt, connection=conn.dbapi_connection, return_type="arrow"
            ).rename_columns({"root_id": "id"})

        selection = ["id", "key"]
        if return_leaf_id:
            selection.append("leaf_id")

        return id_results.select(selection)


def get_parent_clusters_and_leaves(
    resolution: Resolutions,
) -> dict[int, dict[str, list[dict]]]:
    """Query clusters and their leaves for all parent resolutions.

    For a given resolution, find all its parent resolutions and return complete
    cluster compositions.

    * Parent discovery: Queries ResolutionFrom to find all direct parent
        resolutions (level 1)
    * Cluster building: For each parent, runs the full unified query to get all
        cluster assignments with both root and leaf information
    * Aggregation: Collects all leaf nodes belonging to each root cluster across all
        parent resolutions

    Return a dictionary mapping cluster IDs to their complete leaf compositions
    and metadata.
    """
    with MBDB.get_session() as session:
        # Get direct parent resolution IDs
        parent_ids: list[int] = [
            row[0]
            for row in session.execute(
                select(ResolutionFrom.parent)
                .where(ResolutionFrom.child == resolution.resolution_id)
                .where(ResolutionFrom.level == 1)
            ).all()
        ]

        if not parent_ids:
            return {}

        # For each parent, get all cluster assignments
        all_clusters: dict[int, dict] = {}

        for parent_id in parent_ids:
            parent_resolution: Resolutions = session.get(Resolutions, parent_id)
            if parent_resolution is None:
                continue

            parent_assignments: Select = build_unified_query(
                resolution=parent_resolution, get_hashes=True
            )

            for row in session.execute(parent_assignments):
                root_id = row.root_id
                if root_id not in all_clusters:
                    all_clusters[root_id] = {
                        "root_hash": row.root_hash,
                        "leaves": [],
                        "probability": None,
                    }
                all_clusters[root_id]["leaves"].append(
                    {"leaf_id": row.leaf_id, "leaf_hash": row.leaf_hash}
                )

        return all_clusters


def match(
    key: str,
    source: SourceResolutionName,
    targets: list[SourceResolutionName],
    resolution: ResolutionName,
    threshold: int | None = None,
) -> list[Match]:
    """Matches an ID in a source resolution and returns the keys in the targets.

    Given a specific key in a source, find what it matches to in target sources
    through a resolution hierarchy.

    * Target cluster identification: Uses COALESCE priority CTE to determine which
        cluster the input key belongs to at the resolution level
    * Matching leaves discovery: Builds UNION ALL query with branches for:
        * Direct cluster members (source-only case)
        * Members connected through each model resolution in the hierarchy
    * Cross-reference: Joins the target cluster with all possible matching leaves,
        filtering for the requested target sources

    Organises matches by source configuration and returns structured Match objects
    for each target.
    """
    with MBDB.get_session() as session:
        # Get configurations
        source_config: SourceConfigs = get_source_config(source, session)
        truth_resolution: Resolutions | None = session.execute(
            select(Resolutions).where(Resolutions.name == resolution)
        ).scalar()
        if truth_resolution is None:
            raise MatchboxResolutionNotFoundError(name=resolution)

        target_configs: list[SourceConfigs] = [
            get_source_config(target, session) for target in targets
        ]
        target_source_config_ids = [tc.source_config_id for tc in target_configs]

        # Build and execute the match query
        matches_query = _build_match_query(
            key=key,
            source_config_id=source_config.source_config_id,
            target_source_config_ids=target_source_config_ids,
            resolution=truth_resolution,
            threshold=threshold,
        )

        logger.debug(f"Match SQL: \n {compile_sql(matches_query)}")
        matches = session.execute(matches_query).all()

        # Organise matches by source config
        cluster: int | None = None
        matches_by_source_id: dict[int, set] = {}

        for cluster_id, source_config_id_result, key_in_source in matches:
            if cluster is None:
                cluster = cluster_id

            # Skip NULL results from LEFT JOIN (when no target matches)
            if source_config_id_result is not None and key_in_source is not None:
                if source_config_id_result not in matches_by_source_id:
                    matches_by_source_id[source_config_id_result] = set()
                matches_by_source_id[source_config_id_result].add(key_in_source)

        # Build result objects
        result: list[Match] = []
        for target, target_config in zip(targets, target_configs, strict=False):
            match_obj = Match(
                cluster=cluster,
                source=source,
                source_id=matches_by_source_id.get(
                    source_config.source_config_id, set()
                ),
                target=target,
                target_id=matches_by_source_id.get(
                    target_config.source_config_id, set()
                ),
            )
            result.append(match_obj)

        return result



from pathlib import Path

import pandas as pd
from sqlalchemy import create_engine, text
from sqlalchemy.types import TypeEngine


def dataset_parquet_to_local_warehouse(
    parquetpath: Path, schema: str, table: str, dtype: dict[str, TypeEngine]
) -> None:
    """Load dataset parquet file to local warehouse instance.

    Args:
        parquetpath: Path to parquet file.
        schema: Name of schema in local database.
        table: Name of new table in local database.
        dtype: Dictionary of dataset column names and sqlalchemy types.
    """

    user = "warehouse_user"
    password = "warehouse_password"
    host = "localhost"
    database = "warehouse"
    port = 7654

    engine = create_engine(
        f"postgresql+psycopg://{user}:{password}@{host}:{port}/{database}"
    )
    with engine.connect() as connection:
        connection.execute(text(f"CREATE SCHEMA IF NOT EXISTS {schema}"))
        connection.commit()

    dataset = pd.read_parquet(parquetpath, dtype_backend="pyarrow")
    dataset.to_sql(
        name=table,
        con=engine,
        schema=schema,
        if_exists="replace",
        dtype=dtype,
        chunksize=100000,
        index=False,
    )
    engine.dispose()



"""Base classes and utilities for Matchbox database adapters."""

import json
from abc import ABC, abstractmethod
from enum import StrEnum
from typing import TYPE_CHECKING, Any, Literal, Protocol, Self

import boto3
from botocore.exceptions import ClientError
from pyarrow import Table
from pydantic import BaseModel, Field, SecretStr, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from matchbox.common.dtos import Match, Resolution
from matchbox.common.eval import Judgement, ModelComparison
from matchbox.common.graph import (
    ModelResolutionName,
    ResolutionGraph,
    ResolutionName,
    ResolutionType,
    SourceResolutionName,
)
from matchbox.common.logging import LogLevelType

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any


class MatchboxBackends(StrEnum):
    """The available backends for Matchbox."""

    POSTGRES = "postgres"


class MatchboxSnapshot(BaseModel):
    """A snapshot of the Matchbox database."""

    backend_type: MatchboxBackends
    data: Any

    @field_validator("data")
    @classmethod
    def check_serialisable(cls, value: Any) -> Any:
        """Validate that the value can be serialised to JSON."""
        try:
            json.dumps(value)
            return value
        except (TypeError, OverflowError) as e:
            raise ValueError(f"Value is not JSON serialisable: {e}") from e


class MatchboxDatastoreSettings(BaseSettings):
    """Settings specific to the datastore configuration."""

    host: str | None = None
    port: int | None = None
    access_key_id: SecretStr | None = None
    secret_access_key: SecretStr | None = None
    default_region: str | None = None
    cache_bucket_name: str

    def get_client(self) -> S3Client:
        """Returns an S3 client for the datastore.

        Creates S3 buckets if they don't exist.
        """
        kwargs = {
            "endpoint_url": f"http://{self.host}:{self.port}"
            if self.host and self.port
            else None,
            "aws_access_key_id": self.access_key_id.get_secret_value()
            if self.access_key_id
            else None,
            "aws_secret_access_key": self.secret_access_key.get_secret_value()
            if self.secret_access_key
            else None,
            "region_name": self.default_region,
        }

        client: S3Client = boto3.client("s3", **kwargs)

        try:
            client.head_bucket(Bucket=self.cache_bucket_name)
        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":
                client.create_bucket(
                    Bucket=self.cache_bucket_name,
                    CreateBucketConfiguration={
                        "LocationConstraint": self.default_region
                    },
                )
            else:
                raise e

        return client


class MatchboxServerSettings(BaseSettings):
    """Settings for the Matchbox application."""

    model_config = SettingsConfigDict(
        env_prefix="MB__SERVER__",
        env_nested_delimiter="__",
        use_enum_values=True,
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    batch_size: int = Field(default=250_000)
    backend_type: MatchboxBackends
    datastore: MatchboxDatastoreSettings
    task_runner: Literal["api", "celery"]
    redis_uri: str | None
    uploads_expiry_minutes: int | None
    authorisation: bool = False
    public_key: SecretStr | None = Field(default=None)
    log_level: LogLevelType = "INFO"

    @model_validator(mode="after")
    def check_settings(self) -> Self:
        """Check that legal combinations of settings are provided."""
        if self.task_runner == "celery" and self.redis_uri is None:
            raise ValueError("A Redis URI must be set if using Celery as task runner.")
        if self.task_runner == "celery" and self.uploads_expiry_minutes is None:
            raise ValueError(
                "Upload expiration must be set if using Celery as task runner."
            )

        return self


class BackendManager:
    """Manages the Matchbox backend instance and settings."""

    _instance = None
    _settings = None

    @classmethod
    def initialise(cls, settings: "MatchboxServerSettings"):
        """Initialise the backend with the given settings."""
        cls._settings = settings

    @classmethod
    def get_backend(cls) -> "MatchboxDBAdapter":
        """Get the backend instance."""
        if cls._settings is None:
            raise ValueError("BackendManager must be initialized with settings first")

        if cls._instance is None:
            BackendClass = get_backend_class(cls._settings.backend_type)
            cls._instance = BackendClass(cls._settings)
        return cls._instance

    @classmethod
    def get_settings(cls) -> "MatchboxServerSettings":
        """Get the backend settings."""
        if cls._settings is None:
            raise ValueError("BackendManager must be initialized with settings first")
        return cls._settings


def get_backend_settings(
    backend_type: MatchboxBackends,
) -> type[MatchboxServerSettings]:
    """Get the appropriate settings class based on the backend type."""
    if backend_type == MatchboxBackends.POSTGRES:
        from matchbox.server.postgresql import MatchboxPostgresSettings  # noqa: PLC0415

        return MatchboxPostgresSettings
    # Add more backend types here as needed
    else:
        raise ValueError(f"Unsupported backend type: {backend_type}")


def get_backend_class(backend_type: MatchboxBackends) -> type["MatchboxDBAdapter"]:
    """Get the appropriate backend class based on the backend type."""
    if backend_type == MatchboxBackends.POSTGRES:
        from matchbox.server.postgresql import MatchboxPostgres  # noqa: PLC0415

        return MatchboxPostgres
    # Add more backend types here as needed
    else:
        raise ValueError(f"Unsupported backend type: {backend_type}")


def settings_to_backend(settings: MatchboxServerSettings) -> "MatchboxDBAdapter":
    """Create backend adapter with injected settings."""
    BackendClass = get_backend_class(settings.backend_type)
    return BackendClass(settings)


def initialise_matchbox() -> None:
    """Initialise the Matchbox backend based on environment variables."""
    base_settings = MatchboxServerSettings()

    SettingsClass = get_backend_settings(base_settings.backend_type)
    settings = SettingsClass()

    BackendManager.initialise(settings)


class Countable(Protocol):
    """A protocol for objects that can be counted."""

    def count(self) -> int:
        """Counts the number of items in the object."""
        ...


class Listable(Protocol):
    """A protocol for objects that can be listed."""

    def list_all(self) -> list[str]:
        """Lists the items in the object."""
        ...


class ListableAndCountable(Countable, Listable):
    """A protocol for objects that can be counted and listed."""

    pass


class MatchboxDBAdapter(ABC):
    """An abstract base class for Matchbox database adapters."""

    settings: "MatchboxServerSettings"

    sources: ListableAndCountable
    models: Countable
    data: Countable
    clusters: Countable
    creates: Countable
    merges: Countable
    proposes: Countable
    source_resolutions: Countable

    # Retrieval

    @abstractmethod
    def query(
        self,
        source: SourceResolutionName,
        resolution: ResolutionName | None = None,
        threshold: int | None = None,
        return_leaf_id: bool = False,
        limit: int | None = None,
    ) -> Table:
        """Queries the database from an optional point of truth.

        Args:
            source: the `SourceResolutionName` string identifying the source to query
            resolution (optional): the resolution to use for filtering results
                If not specified, will use the source resolution for the queried source
            threshold (optional): the threshold to use for creating clusters
                If None, uses the models' default threshold
                If an integer, uses that threshold for the specified model, and the
                model's cached thresholds for its ancestors
            return_leaf_id (optional): whether to return cluster ID of leaves
            limit (optional): the number to use in a limit clause. Useful for testing

        Returns:
            The resulting matchbox IDs in Arrow format
        """
        ...

    @abstractmethod
    def match(
        self,
        key: str,
        source: SourceResolutionName,
        targets: list[SourceResolutionName],
        resolution: ResolutionName,
        threshold: int | None = None,
    ) -> list[Match]:
        """Matches an ID in a source resolution and returns the keys in the targets.

        Args:
            key: The key to match from the source.
            source: The name of the source resolution.
            targets: The names of the target source resolutions.
            resolution: The name of the resolution to use for matching.
            threshold (optional): the threshold to use for creating clusters
                If None, uses the resolutions' default threshold
                If an integer, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors
                Will use these threshold values instead of the cached thresholds
        """
        ...

    # Resolution management

    @abstractmethod
    def insert_resolution(self, resolution: "Resolution") -> None:
        """Writes a resolution to Matchbox.

        Args:
            resolution: Resolution object with a source or model config

        Raises:
            MatchboxModelConfigError: If the configuration is invalid, such as
                the ModelConfig's resolutions sharing ancestors
        """
        ...

    @abstractmethod
    def get_resolution(
        self, name: ResolutionName, validate: ResolutionType | None = None
    ) -> Resolution:
        """Get a resolution from its name.

        Args:
            name: The name resolution name for the source
            validate: The expected type of the resolution

        Returns:
            A Resolution object
        """
        ...

    @abstractmethod
    def delete_resolution(self, name: ResolutionName, certain: bool) -> None:
        """Delete a resolution from the database.

        Args:
            name: The name of the resolution to delete.
            certain: Whether to delete the model without confirmation.
        """
        ...

    @abstractmethod
    def get_leaf_source_resolutions(self, name: ResolutionName) -> list[Resolution]:
        """Get a list of source configurations queriable from a resolution.

        Args:
            name: Name of the resolution to query.

        Returns:
            List of relevant Resolution objects.
        """
        ...

    @abstractmethod
    def get_resolution_graph(self) -> ResolutionGraph:
        """Get the full resolution graph."""
        ...

    # Data insertion

    @abstractmethod
    def insert_source_data(
        self, name: SourceResolutionName, data_hashes: Table
    ) -> None:
        """Inserts hash data for a source resolution.

        Args:
            name: The name of the source resolution to index.
            data_hashes: The Arrow table with the hash of each data row
        """
        ...

    @abstractmethod
    def insert_model_data(self, name: ModelResolutionName, results: Table) -> None:
        """Inserts results data for a model resolution."""
        ...

    @abstractmethod
    def get_model_data(self, name: ModelResolutionName) -> Table:
        """Get the results for a model resolution."""
        ...

    @abstractmethod
    def set_model_truth(self, name: ModelResolutionName, truth: int) -> None:
        """Sets the truth threshold for this model, changing the default clusters."""
        ...

    @abstractmethod
    def get_model_truth(self, name: ModelResolutionName) -> int:
        """Gets the current truth threshold for this model."""
        ...

    # Data management

    @abstractmethod
    def validate_ids(self, ids: list[int]) -> bool:
        """Validates a list of IDs exist in the database.

        Args:
            ids: A list of IDs to validate.

        Raises:
            MatchboxDataNotFound: If some items don't exist in the target table.
        """
        ...

    @abstractmethod
    def validate_hashes(self, hashes: list[bytes]) -> bool:
        """Validates a list of hashes exist in the database.

        Args:
            hashes: A list of hashes to validate.

        Raises:
            MatchboxDataNotFound: If some items don't exist in the target table.
        """
        ...

    @abstractmethod
    def cluster_id_to_hash(self, ids: list[int]) -> dict[int, bytes | None]:
        """Get a lookup of Cluster hashes from a list of IDs.

        Args:
            ids: A list of IDs to get hashes for.

        Returns:
            A dictionary mapping IDs to hashes.
        """
        ...

    @abstractmethod
    def dump(self) -> MatchboxSnapshot:
        """Dumps the entire database to a snapshot.

        Returns:
            A MatchboxSnapshot object of type "postgres" with the database's
                current state.
        """
        ...

    @abstractmethod
    def drop(self, certain: bool) -> None:
        """Hard clear the database by dropping all tables and re-creating.

        Args:
            certain: Whether to drop the database without confirmation.
        """
        ...

    @abstractmethod
    def clear(self, certain: bool) -> None:
        """Soft clear the database by deleting all rows but retaining tables.

        Args:
            certain: Whether to delete the database without confirmation.
        """
        ...

    @abstractmethod
    def restore(self, snapshot: MatchboxSnapshot) -> None:
        """Restores the database from a snapshot.

        Args:
            snapshot: A MatchboxSnapshot object of type "postgres" with the
                database's state

        Raises:
            TypeError: If the snapshot is not compatible with PostgreSQL
        """
        ...

    # User management

    @abstractmethod
    def login(self, user_name: str) -> int:
        """Receives a user name and returns user ID."""

    # Evaluation management

    @abstractmethod
    def insert_judgement(self, judgement: Judgement) -> None:
        """Adds an evaluation judgement to the database.

        Args:
            judgement: representation of the proposed clusters.
        """
        ...

    @abstractmethod
    def get_judgements(self) -> tuple[Table, Table]:
        """Retrieves all evaluation judgements.

        Returns:
            Two PyArrow tables with the judgments and their expansion.
            See `matchbox.common.arrow` for information on the schema.
        """
        ...

    @abstractmethod
    def compare_models(self, resolutions: list[ModelResolutionName]) -> ModelComparison:
        """Compare metrics of models based on evaluation data.

        Args:
            resolutions: List of names of model resolutions to be compared.

        Returns:
            A model comparison object, listing metrics for each model.
        """
        ...

    @abstractmethod
    def sample_for_eval(
        self, n: int, resolution: ModelResolutionName, user_id: int
    ) -> Table:
        """Sample a cluster to validate.

        Args:
            n: Number of clusters to sample
            resolution: Name of resolution from which to sample
            user_id: ID of user requesting the sample

        Returns:
            An Arrow table with the same schema as returned by `query()`
        """
        ...



"""Renamed primary keys to just keys.

Revision ID: e4122bdf9b0d
Revises: 83b134a86713
Create Date: 2025-05-13 13:10:12.346864

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e4122bdf9b0d"
down_revision: str | None = "83b134a86713"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # Rename column in pk_space table
    op.alter_column(
        "pk_space",
        "next_cluster_source_pk_id",
        new_column_name="next_cluster_keys_id",
        schema="mb",
    )

    # Rename cluster_source_pks table to cluster_keys
    op.rename_table("cluster_source_pks", "cluster_keys", schema="mb")

    # Rename columns in cluster_keys table
    op.alter_column("cluster_keys", "pk_id", new_column_name="key_id", schema="mb")
    op.alter_column("cluster_keys", "source_pk", new_column_name="key", schema="mb")

    # Rename constraint in cluster_keys table
    op.drop_constraint("unique_pk_source", "cluster_keys", schema="mb")
    op.create_unique_constraint(
        "unique_keys_source",
        "cluster_keys",
        ["key_id", "source_config_id"],
        schema="mb",
    )

    # Rename indexes (they need to be dropped and recreated with new names)
    op.drop_index(
        "ix_cluster_source_pks_cluster_id", table_name="cluster_keys", schema="mb"
    )
    op.drop_index(
        "ix_cluster_source_pks_source_pk", table_name="cluster_keys", schema="mb"
    )

    op.create_index(
        "ix_cluster_keys_cluster_id",
        "cluster_keys",
        ["cluster_id"],
        unique=False,
        schema="mb",
    )
    op.create_index(
        "ix_cluster_keys_keys",
        "cluster_keys",
        ["key"],
        unique=False,
        schema="mb",
    )

    # Rename column in source_configs table
    op.alter_column("source_configs", "db_pk", new_column_name="key_field", schema="mb")


def downgrade() -> None:
    """Downgrade schema."""
    # Rename column in source_configs table back
    op.alter_column("source_configs", "key_field", new_column_name="db_pk", schema="mb")

    # Rename indexes back
    op.drop_index("ix_cluster_keys_cluster_id", table_name="cluster_keys", schema="mb")
    op.drop_index("ix_cluster_keys_keys", table_name="cluster_keys", schema="mb")

    op.create_index(
        "ix_cluster_source_pks_cluster_id",
        "cluster_keys",
        ["cluster_id"],
        unique=False,
        schema="mb",
    )
    op.create_index(
        "ix_cluster_source_pks_source_pk",
        "cluster_keys",
        ["key"],  # Note: column is still named 'key' at this point
        unique=False,
        schema="mb",
    )

    # Rename constraint back
    op.drop_constraint("unique_keys_source", "cluster_keys", schema="mb")
    op.create_unique_constraint(
        "unique_pk_source", "cluster_keys", ["key_id", "source_config_id"], schema="mb"
    )

    # Rename columns back in cluster_keys table
    op.alter_column("cluster_keys", "key", new_column_name="source_pk", schema="mb")
    op.alter_column("cluster_keys", "key_id", new_column_name="pk_id", schema="mb")

    # Rename cluster_keys table back to cluster_source_pks
    op.rename_table("cluster_keys", "cluster_source_pks", schema="mb")

    # Rename column in pk_space table back to next_cluster_source_pk_id
    op.alter_column(
        "pk_space",
        "next_cluster_keys_id",
        new_column_name="next_cluster_source_pk_id",
        schema="mb",
    )



"""Simplify resolution naming and hashing.

Revision ID: 83b134a86713
Revises: 95c0b5c23446
Create Date: 2025-05-12 17:11:45.330677

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from matchbox.common.hash import hash_values

# revision identifiers, used by Alembic.
revision: str = "83b134a86713"
down_revision: str | None = "95c0b5c23446"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def calculate_resolution_hashes(
    resolutions: dict[int, tuple[str, str]],
    relationships: list[tuple[int, int]],
) -> list[tuple[int, bytes]]:
    """Calculate resolution_hash for each resolution based on its type and dependencies.

    Args:
        resolutions: Dict mapping resolution_id -> (name, type)
        relationships: List of (child, parent) tuples for direct parent relationships

    Returns:
        List of (resolution_id, calculated_hash) tuples
    """
    # Build parent relationships
    parents: dict[int, list[int]] = {}
    for child, parent in relationships:
        parents.setdefault(child, []).append(parent)

    # Calculate hashes recursively
    cache: dict[int, bytes] = {}

    def get_hash(rid: int) -> bytes:
        if rid not in cache:
            name, rtype = resolutions[rid]
            if rtype == "dataset":
                cache[rid] = hash_values(name)
            else:
                parent_hashes = sorted(get_hash(p) for p in parents.get(rid, []))
                cache[rid] = hash_values(*parent_hashes, bytes(name, encoding="utf-8"))
        return cache[rid]

    return [(rid, get_hash(rid)) for rid in resolutions]


def upgrade() -> None:
    """Upgrade schema."""
    # Drop the old check constraint
    op.drop_constraint(
        "resolution_type_constraints", "resolutions", schema="mb", type_="check"
    )

    # Update any 'dataset' types to 'source' types
    op.execute(
        """
        UPDATE mb.resolutions 
        SET type = 'source'
        WHERE type = 'dataset'
        """
    )

    # Create the new check constraint with updated type values
    op.create_check_constraint(
        "resolution_type_constraints",
        "resolutions",
        "type IN ('model', 'source', 'human')",
        schema="mb",
    )

    # Add the new hash column
    op.add_column(
        "resolutions", sa.Column("hash", postgresql.BYTEA(), nullable=True), schema="mb"
    )

    # Populate the new hash column based on type
    op.execute(
        """
        UPDATE mb.resolutions 
        SET hash = CASE 
            WHEN type = 'model' THEN resolution_hash
            WHEN type = 'source' THEN content_hash
            ELSE NULL
        END
        """
    )

    # Drop the unique constraint on resolution_hash
    op.drop_constraint(
        "resolutions_hash_key", "resolutions", schema="mb", type_="unique"
    )

    # Drop the old hash columns
    op.drop_column("resolutions", "resolution_hash", schema="mb")
    op.drop_column("resolutions", "content_hash", schema="mb")

    # Drop the resolution_name column from source_configs
    op.drop_column("source_configs", "resolution_name", schema="mb")


def downgrade() -> None:
    """Downgrade schema."""
    # Add back the resolution_name column to source_configs as nullable first
    op.add_column(
        "source_configs",
        sa.Column("resolution_name", sa.TEXT(), autoincrement=False, nullable=True),
        schema="mb",
    )

    # Populate source_configs.resolution_name with resolutions.name
    # Use the correct column names based on the ORM
    op.execute(
        """
        UPDATE mb.source_configs 
        SET resolution_name = r.name
        FROM mb.resolutions r
        WHERE source_configs.resolution_id = r.resolution_id
        """
    )

    # Now make the column NOT NULL after it's been populated
    op.alter_column("source_configs", "resolution_name", nullable=False, schema="mb")

    # Add back the old hash columns
    op.add_column(
        "resolutions",
        sa.Column(
            "content_hash", postgresql.BYTEA(), autoincrement=False, nullable=True
        ),
        schema="mb",
    )
    op.add_column(
        "resolutions",
        sa.Column(
            "resolution_hash", postgresql.BYTEA(), autoincrement=False, nullable=True
        ),
        schema="mb",
    )

    # Load data for hash calculation
    connection = op.get_bind()

    # Get all resolutions
    resolutions = {}
    for rid, name, rtype in connection.execute(
        sa.text("SELECT resolution_id, name, type FROM mb.resolutions")
    ):
        resolutions[rid] = (name, rtype)

    # Get parent relationships
    relationships = list(
        connection.execute(
            sa.text("SELECT child, parent FROM mb.resolution_from WHERE level = 1")
        )
    )

    # Calculate new hashes
    hash_results = calculate_resolution_hashes(resolutions, relationships)

    # Update resolution_hash with calculated values
    for resolution_id, calculated_hash in hash_results:
        connection.execute(
            sa.text(
                "UPDATE mb.resolutions SET resolution_hash = :hash "
                "WHERE resolution_id = :id"
            ),
            {"hash": calculated_hash, "id": resolution_id},
        )

    # Set content_hash from hash for sources only
    op.execute(
        """
        UPDATE mb.resolutions 
        SET content_hash = CASE 
            WHEN type = 'source' THEN hash
            ELSE NULL
        END
        """
    )

    # Now make resolution_hash NOT NULL after it's been populated
    op.alter_column("resolutions", "resolution_hash", nullable=False, schema="mb")

    # Recreate the unique constraint
    op.create_unique_constraint(
        "resolutions_hash_key", "resolutions", ["resolution_hash"], schema="mb"
    )

    # Drop the new hash column
    op.drop_column("resolutions", "hash", schema="mb")

    # Drop the new check constraint
    op.drop_constraint(
        "resolution_type_constraints", "resolutions", schema="mb", type_="check"
    )

    # Update any 'source' types back to 'dataset' types
    op.execute(
        """
        UPDATE mb.resolutions 
        SET type = 'dataset'
        WHERE type = 'source'
        """
    )

    # Recreate the old check constraint with original type values
    op.create_check_constraint(
        "resolution_type_constraints",
        "resolutions",
        "type IN ('model', 'dataset', 'human')",
        schema="mb",
    )



"""A deduplication methodology based on a deterministic set of conditions."""

import duckdb
import polars as pl
from pydantic import Field

from matchbox.client.models.dedupers.base import Deduper, DeduperSettings


class NaiveSettings(DeduperSettings):
    """A data class to enforce the Naive deduper's settings dictionary shape."""

    unique_fields: list[str] = Field(
        description="A list of fields that will form a unique, deduplicated record"
    )


class NaiveDeduper(Deduper):
    """A simple deduper that deduplicates based on a set of boolean conditions."""

    settings: NaiveSettings

    _id_dtype: pl.DataType

    def prepare(self, data: pl.DataFrame) -> None:
        """Prepare the deduper for deduplication."""
        pass

    def dedupe(self, data: pl.DataFrame) -> pl.DataFrame:
        """Deduplicate the dataframe."""
        self._id_dtype = data[self.settings.id].dtype

        df = data.clone()

        join_clause = []
        for field in self.settings.unique_fields:
            join_clause.append(f"l.{field} = r.{field}")
        join_clause_compiled = " and ".join(join_clause)

        # Generate a key to remove row self-match but ALLOW true duplicate
        # rows where all data items are identical in the source
        df = df.with_row_index("_unique_e4003b")

        # We also need to suppress raw.left_id = raw.right_id in cases where
        # we're deduplicating an unnested array of primary keys
        sql = f"""
            select distinct on (list_sort([raw.left_id, raw.right_id]))
                raw.left_id,
                raw.right_id,
                1.0 as probability
            from (
                select
                    l.{self.settings.id} as left_id,
                    r.{self.settings.id} as right_id
                from
                    df l
                inner join df r on
                    (
                        {join_clause_compiled}
                    ) and
                        l._unique_e4003b != r._unique_e4003b
            ) raw
                where raw.left_id != raw.right_id;
        """

        return (
            duckdb.sql(sql)
            .pl()
            .with_columns(
                [
                    pl.col("left_id").cast(self._id_dtype),
                    pl.col("right_id").cast(self._id_dtype),
                    pl.col("probability").cast(pl.Float32),
                ]
            )
        )



import json
from datetime import datetime
from unittest.mock import Mock, patch

import polars as pl
import pyarrow as pa
import pytest
from httpx import Response
from polars.testing import assert_frame_equal
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.models import Model
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.client.sources import Source
from matchbox.common.arrow import SCHEMA_QUERY, table_to_buffer
from matchbox.common.dtos import BackendResourceType, Match, NotFoundError
from matchbox.common.exceptions import (
    MatchboxEmptyServerResponse,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.factories.sources import source_factory, source_from_tuple


@patch.object(Source, "run")
@patch.object(Model, "run")
@patch.object(Source, "sync")
@patch.object(Model, "sync")
def test_dag_run_and_sync(
    model_sync_mock: Mock,
    source_sync_mock: Mock,
    model_run_mock: Mock,
    source_run_mock: Mock,
    sqlite_warehouse: Engine,
):
    """A legal DAG can be built and run."""
    # Set up constituents
    foo_tkit = source_factory(name="foo", engine=sqlite_warehouse).write_to_location()
    bar_tkit = source_factory(name="bar", engine=sqlite_warehouse).write_to_location()
    baz_tkit = source_factory(name="baz", engine=sqlite_warehouse).write_to_location()

    dag = DAG(name="dag")

    # Structure: sources can be added
    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    baz = dag.source(**baz_tkit.into_dag())

    assert set(dag.nodes.keys()) == {foo.name, bar.name, baz.name}

    # Structure: sources can be deduped
    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )

    # Structure:
    # - sources can be passed directly to linkers
    # - or, linkers can take dedupers
    foo_bar = d_foo.query(foo).linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Structure: linkers can take other linkers
    foo_bar_baz = foo_bar.query(foo, bar, baz).linker(
        baz.query(),
        name="foo_bar_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    assert set(dag.nodes.keys()) == {
        foo.name,
        bar.name,
        baz.name,
        d_foo.name,
        foo_bar.name,
        foo_bar_baz.name,
    }

    # Run DAG
    dag.run_and_sync()

    assert source_run_mock.call_count == 3
    assert source_sync_mock.call_count == 3
    assert model_run_mock.call_count == 3
    assert model_sync_mock.call_count == 3


def test_dags_missing_dependency(sqlite_warehouse: Engine):
    """Steps cannot be added before their dependencies."""
    dag = DAG("collection")
    foo = source_factory(name="foo", engine=sqlite_warehouse, dag=dag).source

    with pytest.raises(ValueError, match="not added to DAG"):
        dag.model(
            left_query=foo.query(),
            name="d_foo",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Failure leads to no dags being added
    assert not len(dag.nodes)
    assert not len(dag.graph)


def test_mixing_dags_fails(sqlite_warehouse: Engine):
    """Cannot reference a different DAG when adding a step."""
    dag = DAG("collection")
    dag2 = DAG("collection")

    foo_tkit = source_factory(name="foo", engine=sqlite_warehouse)
    foo = dag.source(**foo_tkit.into_dag())

    with pytest.raises(ValueError, match="mix DAGs"):
        # Different DAG in input
        dag2.model(
            left_query=foo.query(),
            name="d_foo",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Failure leads to no dags being added
    assert not len(dag2.nodes)
    assert not len(dag2.graph)


def test_dag_name_clash(sqlite_warehouse: Engine):
    """Names across sources and steps must be unique."""
    dag = DAG("collection")

    foo_tkit = source_factory(name="foo", engine=sqlite_warehouse)
    bar_tkit = source_factory(name="bar", engine=sqlite_warehouse)

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())

    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )

    with pytest.raises(ValueError, match="already taken"):
        bar.query().deduper(
            name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
        )

    # DAG is not modified by failed attempt
    assert dag.nodes["d_foo"] == d_foo
    # We didn't overwrite d_foo's dependencies
    assert dag.graph["d_foo"] == [foo.name]


def test_dag_disconnected(sqlite_warehouse: Engine):
    """Nodes cannot be disconnected."""
    dag = DAG("collection")

    foo_tkit = source_factory(name="foo", engine=sqlite_warehouse)
    bar_tkit = source_factory(name="bar", engine=sqlite_warehouse)

    dag.source(**foo_tkit.into_dag())
    dag.source(**bar_tkit.into_dag())

    with pytest.raises(ValueError, match="disconnected"):
        dag.run_and_sync()


def test_dag_draw(sqlite_warehouse: Engine):
    """Test that the draw method produces a correct string representation of the DAG."""
    # Set up a simple DAG
    foo_tkit = source_factory(name="foo", engine=sqlite_warehouse).write_to_location()
    bar_tkit = source_factory(name="bar", engine=sqlite_warehouse).write_to_location()
    baz_tkit = source_factory(name="baz", engine=sqlite_warehouse).write_to_location()

    dag = DAG(name="default")

    # Structure: sources can be added
    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    baz = dag.source(**baz_tkit.into_dag())

    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )

    foo_bar = d_foo.query(foo).linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Structure: linkers can take other linkers
    foo_bar_baz = foo_bar.query(foo, bar, baz).linker(
        baz.query(),
        name="foo_bar_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Prepare the DAG and draw it

    # Test 1: Drawing without timestamps (original behavior)
    tree_str = dag.draw()

    # Verify the structure
    lines = tree_str.strip().split("\n")

    # The root node should be first
    assert lines[0] == "foo_bar_baz"

    # Check that all nodes are present
    node_names = [
        foo.name,
        bar.name,
        baz.name,
        d_foo.name,
        foo_bar.name,
        foo_bar_baz.name,
    ]

    for node in node_names:
        # Either the node name is at the start of a line or after the tree characters
        node_present = any(line.endswith(node) for line in lines)
        assert node_present, f"Node {node} not found in the tree representation"

    # Check that tree has correct formatting with tree characters
    tree_chars = ["└──", "├──", "│"]
    has_tree_chars = any(char in tree_str for char in tree_chars)
    assert has_tree_chars, (
        "Tree representation doesn't use expected formatting characters"
    )

    # Test 2: Drawing with timestamps (status indicators)
    # Set d_foo as processing and foo_bar as completed
    start_time = datetime.now()
    doing = "d_foo"
    foo_bar.last_run = datetime.now()

    # Draw the DAG with status indicators
    tree_str_with_status = dag.draw(start_time=start_time, doing=doing)
    status_lines = tree_str_with_status.strip().split("\n")

    # Verify status indicators are present
    status_indicators = ["✅", "🔄", "⏸️"]
    assert any(indicator in tree_str_with_status for indicator in status_indicators)

    # Check specific statuses: foo_bar done, d_foo working, others awaiting
    for line in status_lines:
        name = line.split()[-1]
        if name == "foo_bar":
            assert "✅" in line
        elif name == "d_foo":
            assert "🔄" in line
        elif name in [foo.name, bar.name, baz.name]:
            assert "⏸️" in line

    # Test 3: Check that node names are still present with status indicators
    for node in node_names:
        node_present = any(node in line for line in status_lines)
        assert node_present, (
            f"Node {node} not found in the tree representation with status indicators"
        )

    # Test 4: Drawing with skipped nodes
    skipped_nodes = [foo.name, d_foo.name]
    tree_str_with_skipped = dag.draw(
        start_time=start_time, doing=doing, skipped=skipped_nodes
    )
    skipped_lines = tree_str_with_skipped.strip().split("\n")

    # Check that skipped nodes have the skipped indicator
    for line in skipped_lines:
        name = line.split()[-1]
        if any(name == skipped for skipped in skipped_nodes):
            assert "⏭️" in line

    # Test all status indicators together
    doing = "foo_bar_baz"
    tree_str_all_statuses = dag.draw(
        start_time=start_time, doing=doing, skipped=skipped_nodes
    )
    assert all(
        indicator in tree_str_all_statuses for indicator in ["✅", "🔄", "⏸️", "⏭️"]
    )


# Lookups


def test_extract_lookup(
    sqlite_warehouse: Engine,
    sqlite_in_memory_warehouse: Engine,
    matchbox_api: MockRouter,
):
    """Entire lookup can be extracted from DAG."""
    # Make dummy data

    foo = source_from_tuple(
        name="foo",
        location_name="sqlite",
        engine=sqlite_warehouse,
        data_keys=[1, 2, 3],
        data_tuple=({"col": 0}, {"col": 1}, {"col": 2}),
    ).write_to_location()
    bar = source_from_tuple(
        name="bar",
        location_name="sqlite_memory",
        engine=sqlite_in_memory_warehouse,
        data_keys=["a", "b", "c"],
        data_tuple=({"col": 10}, {"col": 11}, {"col": 12}),
    ).write_to_location()

    dag = DAG("companies")
    dag.source(**foo.into_dag()).query().linker(
        dag.source(**bar.into_dag()).query(),
        name="root",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Because of FULL OUTER JOIN, we expect some values to be null, and some explosions
    expected_foo_bar_mapping = pl.DataFrame(
        [
            {"id": 1, "foo_key": "1", "bar_key": "a"},
            {"id": 2, "foo_key": "2", "bar_key": None},
            {"id": 3, "foo_key": "3", "bar_key": "b"},
            {"id": 3, "foo_key": "3", "bar_key": "c"},
        ]
    )

    # When selecting single source, we won't explode
    expected_foo_mapping = expected_foo_bar_mapping.select(["id", "foo_key"]).unique()

    # Mock API
    matchbox_api.get("/resolutions/root/sources").mock(
        return_value=Response(
            200,
            json=[
                foo.source.to_resolution().model_dump(mode="json"),
                bar.source.to_resolution().model_dump(mode="json"),
            ],
        )
    )

    matchbox_api.get("/query", params={"source": "foo"}).mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"id": 1, "key": "1"},
                        {"id": 2, "key": "2"},
                        {"id": 3, "key": "3"},
                    ],
                    schema=SCHEMA_QUERY,
                )
            ).read(),
        )
    )

    matchbox_api.get("/query", params={"source": "bar"}).mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"id": 1, "key": "a"},
                        {"id": 3, "key": "b"},
                        {"id": 3, "key": "c"},
                    ],
                    schema=SCHEMA_QUERY,
                )
            ).read(),
        )
    )

    # Case 0: No sources are found
    with pytest.raises(MatchboxResolutionNotFoundError):
        dag.extract_lookup(source_filter=["nonexistent"])

    with pytest.raises(MatchboxResolutionNotFoundError):
        dag.extract_lookup(location_names=["nonexistent"])

    # Case 1: Retrieve single table
    # With URI filter
    foo_mapping = dag.extract_lookup(location_names=["sqlite"])

    assert_frame_equal(
        pl.from_arrow(foo_mapping),
        expected_foo_mapping,
        check_row_order=False,
        check_column_order=False,
    )

    # With source filter
    foo_mapping = dag.extract_lookup(source_filter="foo")

    assert_frame_equal(
        pl.from_arrow(foo_mapping),
        expected_foo_mapping,
        check_row_order=False,
        check_column_order=False,
    )

    # With both filters
    foo_mapping = dag.extract_lookup(source_filter="foo", location_names="sqlite")

    assert_frame_equal(
        pl.from_arrow(foo_mapping),
        expected_foo_mapping,
        check_row_order=False,
        check_column_order=False,
    )

    # Case 2: Retrieve multiple tables
    # With no filter
    foo_bar_mapping = dag.extract_lookup()

    assert_frame_equal(
        pl.from_arrow(foo_bar_mapping),
        expected_foo_bar_mapping,
        check_row_order=False,
        check_column_order=False,
    )

    # With source filter
    foo_bar_mapping = dag.extract_lookup(source_filter=["foo", "bar"])

    assert_frame_equal(
        pl.from_arrow(foo_bar_mapping),
        expected_foo_bar_mapping,
        check_row_order=False,
        check_column_order=False,
    )


def test_lookup_key_ok(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """The DAG can map between single keys."""
    # Set up dummy data
    foo_testkit = source_factory(
        engine=sqlite_warehouse, name="foo"
    ).write_to_location()
    bar_testkit = source_factory(
        engine=sqlite_warehouse, name="bar"
    ).write_to_location()
    baz_testkit = source_factory(
        engine=sqlite_warehouse, name="baz"
    ).write_to_location()

    dag = DAG("companies")
    foo = dag.source(**foo_testkit.into_dag())
    bar = dag.source(**bar_testkit.into_dag())
    baz = dag.source(**baz_testkit.into_dag())

    foo.query().linker(
        bar.query(),
        name="linker1",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    ).query(foo, bar).linker(
        baz.query(),
        name="linker2",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    mock_match1 = Match(
        cluster=1, source="foo", source_id={"a"}, target="bar", target_id={"b"}
    )
    mock_match2 = Match(
        cluster=1, source="foo", source_id={"a"}, target="baz", target_id={"b"}
    )
    # The standard JSON serialiser does not handle Pydantic objects
    serialised_matches = json.dumps(
        [m.model_dump() for m in [mock_match1, mock_match2]]
    )

    matchbox_api.get("/match").mock(
        return_value=Response(200, content=serialised_matches)
    )

    # Use lookup function
    matches = dag.lookup_key(from_source="foo", to_sources=["bar", "baz"], key="pk1")

    # Verify results
    assert matches == {foo.name: ["a"], bar.name: ["b"], baz.name: ["b"]}


def test_lookup_key_404_source(matchbox_api: MockRouter):
    """Key lookup throws a resolution not found error."""
    # Set up dummy data
    source_testkit = source_factory(name="source")
    target_testkit = source_factory(name="target")

    dag = DAG("companies")
    dag.source(**source_testkit.into_dag()).query().linker(
        dag.source(**target_testkit.into_dag()).query(),
        name="root",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    matchbox_api.get("/match").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Resolution 42 not found",
                entity=BackendResourceType.RESOLUTION,
            ).model_dump(),
        )
    )

    # Use match function
    with pytest.raises(MatchboxResolutionNotFoundError, match="42"):
        dag.lookup_key(from_source="source", to_sources=["target"], key="pk1")


def test_lookup_key_no_matches(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Key lookup raises MatchboxEmptyServerResponse when no matches are found."""
    # Set up dummy data
    source_testkit = source_factory(
        engine=sqlite_warehouse, name="source"
    ).write_to_location()
    target_testkit = source_factory(
        engine=sqlite_warehouse, name="target"
    ).write_to_location()

    dag = DAG("companies")
    dag.source(**source_testkit.into_dag()).query().linker(
        dag.source(**target_testkit.into_dag()).query(),
        name="root",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Mock empty match results
    matchbox_api.get("/match").mock(return_value=Response(200, content="[]"))

    # Test that empty match results raise MatchboxEmptyServerResponse
    with pytest.raises(
        MatchboxEmptyServerResponse, match="The match operation returned no data"
    ):
        dag.lookup_key(from_source="source", to_sources=["target"], key="pk1")



"""Module to load client settings from env file."""

from pydantic import SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict

from matchbox.common.exceptions import MatchboxClientSettingsException


class ClientSettings(BaseSettings):
    api_root: str
    timeout: float | None = None
    private_key: SecretStr | None = None
    retry_delay: int = 15
    default_warehouse: str | None = None
    jwt: str | None = None
    user: str | None = None
    batch_size: int = 250_000

    model_config = SettingsConfigDict(
        extra="ignore",
        env_prefix="MB__CLIENT__",
        env_nested_delimiter="__",
        env_file=".env",
        env_file_encoding="utf-8",
    )


try:
    settings = ClientSettings()
except ValueError as e:
    raise MatchboxClientSettingsException from e



"""Add content hash to Resolutions.

Revision ID: f3c9279437f4
Revises: b694eb292dea
Create Date: 2025-04-22 08:36:37.054592

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "f3c9279437f4"
down_revision: str | None = "b694eb292dea"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.add_column(
        "resolutions",
        sa.Column("content_hash", postgresql.BYTEA(), nullable=True),
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_column("resolutions", "content_hash", schema="mb")



from importlib.metadata import version
from io import BytesIO
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, call, patch

import pyarrow as pa
import pyarrow.parquet as pq
from fastapi.testclient import TestClient

from matchbox.client.authorisation import (
    generate_json_web_token,
)
from matchbox.common.arrow import SCHEMA_QUERY, table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    LoginAttempt,
    LoginResult,
    Match,
    OKMessage,
    UploadStage,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.factories.sources import source_factory
from matchbox.common.graph import ResolutionGraph

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any


# General


def test_healthcheck(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test the healthcheck endpoint."""
    test_client, _, _ = api_client_and_mocks
    response = test_client.get("/health")
    assert response.status_code == 200
    response = OKMessage.model_validate(response.json())
    assert response.status == "OK"
    assert response.version == version("matchbox-db")


def test_login(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test the login endpoint."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.login = Mock(return_value=1)

    response = test_client.post(
        "/login", json=LoginAttempt(user_name="alice").model_dump()
    )

    assert response.status_code == 200
    response = LoginResult.model_validate(response.json())
    assert response.user_id == 1


# We can patch BackgroundTasks as the api_client_and_mocks fixture
# ensures the API runs the task (not Celery)
@patch("matchbox.server.api.main.BackgroundTasks.add_task")
def test_upload(
    mock_add_task: Mock,
    s3: S3Client,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test uploading a file, happy path."""
    # Setup
    test_client, mock_backend, mock_tracker = api_client_and_mocks

    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.settings.datastore.cache_bucket_name = "test-bucket"
    mock_backend.insert_resolution = Mock(return_value=None)
    mock_backend.insert_source_data = Mock(return_value=None)
    s3.create_bucket(
        Bucket="test-bucket",
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )

    source_testkit = source_factory()

    update_id = mock_tracker.add_source(source_testkit.source.to_resolution())

    # Make request with mocked background task
    response = test_client.post(
        f"/upload/{update_id}",
        files={
            "file": (
                "hashes.parquet",
                table_to_buffer(source_testkit.data_hashes),
                "application/octet-stream",
            ),
        },
    )

    # Validate response
    assert UploadStatus.model_validate(response.json())
    assert response.status_code == 202, response.json()
    assert (
        response.json()["stage"] == UploadStage.QUEUED
    )  # Updated to check for queued status
    # Check both status updates were called in correct order
    assert mock_tracker.update.call_args_list == [
        call(update_id, UploadStage.QUEUED),
    ]
    mock_backend.insert_resolution.assert_not_called()  # Index happens in background
    mock_backend.insert_source_data.assert_not_called()
    mock_add_task.assert_called_once()  # Verify task was queued


# We can patch BackgroundTasks as the api_client_and_mocks fixture
# ensures the API runs the task (not Celery)
@patch("matchbox.server.api.main.BackgroundTasks.add_task")
def test_upload_wrong_schema(
    mock_add_task: Mock,
    s3: S3Client,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test uploading a file with wrong schema."""
    test_client, mock_backend, mock_tracker = api_client_and_mocks
    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.settings.datastore.cache_bucket_name = "test-bucket"

    # Create source with results schema instead of index
    source_testkit = source_factory()
    update_id = mock_tracker.add_source(source_testkit.source.to_resolution())

    response = test_client.post(
        f"/upload/{update_id}",
        files={
            "file": (
                "hashes.parquet",
                table_to_buffer(source_testkit.data),
                "application/octet-stream",
            ),
        },
    )

    # Should fail before task starts
    assert response.status_code == 400
    assert response.json()["stage"] == UploadStage.FAILED
    assert "schema mismatch" in response.json()["details"].lower()
    mock_add_task.assert_not_called()


def test_upload_status_check(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test checking status of an upload using the status endpoint."""
    test_client, _, mock_tracker = api_client_and_mocks
    source_testkit = source_factory()
    update_id = mock_tracker.add_source(source_testkit.source.to_resolution())
    mock_tracker.update(update_id, UploadStage.PROCESSING)
    mock_tracker.reset_mock()

    response = test_client.get(f"/upload/{update_id}/status")

    # Should return current status
    assert response.status_code == 200
    assert response.json()["stage"] == UploadStage.PROCESSING
    mock_tracker.update.assert_not_called()


def test_upload_already_processing(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test attempting to upload when status is already processing."""
    test_client, _, mock_tracker = api_client_and_mocks
    source_testkit = source_factory()
    update_id = mock_tracker.add_source(source_testkit.source.to_resolution())
    mock_tracker.update(update_id, UploadStage.PROCESSING)

    response = test_client.post(
        f"/upload/{update_id}",
        files={"file": ("test.parquet", b"dummy data", "application/octet-stream")},
    )

    # Should return 400 with current status
    assert response.status_code == 400
    assert response.json()["stage"] == UploadStage.PROCESSING


def test_upload_already_queued(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test attempting to upload when status is already queued."""
    test_client, _, mock_tracker = api_client_and_mocks
    source_testkit = source_factory()
    update_id = mock_tracker.add_source(source_testkit.source.to_resolution())
    mock_tracker.update(update_id, UploadStage.QUEUED)

    response = test_client.post(
        f"/upload/{update_id}",
        files={"file": ("test.parquet", b"dummy data", "application/octet-stream")},
    )

    # Should return 400 with current status
    assert response.status_code == 400
    assert response.json()["stage"] == UploadStage.QUEUED


def test_status_check_not_found(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test checking status for non-existent upload ID."""
    test_client, _, mock_tracker = api_client_and_mocks
    mock_tracker.get.return_value = None

    response = test_client.get("/upload/nonexistent-id/status")

    assert response.status_code == 400
    assert response.json()["stage"] == "unknown"
    assert "not found or expired" in response.json()["details"].lower()


# Retrieval


def test_query(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.query = Mock(
        return_value=pa.Table.from_pylist(
            [
                {"keys": "a", "id": 1},
                {"keys": "b", "id": 2},
            ],
            schema=SCHEMA_QUERY,
        )
    )

    response = test_client.get(
        "/query",
        params={"source": "foo", "return_leaf_id": False},
    )

    buffer = BytesIO(response.content)
    table = pq.read_table(buffer)

    assert response.status_code == 200
    assert table.schema.equals(SCHEMA_QUERY)


def test_query_404_resolution(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.query = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get(
        "/query",
        params={"source": "foo", "resolution": "bar", "return_leaf_id": True},
    )

    assert response.status_code == 404


def test_query_404_source(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.query = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get(
        "/query",
        params={"source": "foo", "return_leaf_id": True},
    )

    assert response.status_code == 404


def test_match(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_matches = [
        Match(
            cluster=1,
            source="foo",
            source_id={"1"},
            target="bar",
            target_id={"a"},
        )
    ]
    mock_backend.match = Mock(return_value=mock_matches)

    response = test_client.get(
        "/match",
        params={
            "targets": "foo",
            "source": "bar",
            "key": 1,
            "resolution": "res",
            "threshold": 50,
        },
    )

    assert response.status_code == 200
    [Match.model_validate(m) for m in response.json()]


def test_match_404_resolution(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.match = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get(
        "/match",
        params={
            "targets": ["foo"],
            "source": "bar",
            "key": 1,
            "resolution": "res",
        },
    )

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


def test_match_404_source(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.match = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get(
        "/match",
        params={
            "targets": ["foo"],
            "source": "bar",
            "key": 1,
            "resolution": "res",
        },
    )

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


# Admin


def test_count_all_backend_items(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test the unparameterised entity counting endpoint."""
    test_client, mock_backend, _ = api_client_and_mocks
    entity_counts = {
        "sources": 1,
        "models": 2,
        "data": 3,
        "clusters": 4,
        "creates": 5,
        "merges": 6,
        "proposes": 7,
    }
    for e, c in entity_counts.items():
        mock_e = Mock()
        mock_e.count = Mock(return_value=c)
        setattr(mock_backend, e, mock_e)

    response = test_client.get("/database/count")
    assert response.status_code == 200
    assert response.json() == {"entities": entity_counts}


def test_count_backend_item(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test the parameterised entity counting endpoint."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.models.count = Mock(return_value=20)

    response = test_client.get("/database/count", params={"entity": "models"})
    assert response.status_code == 200
    assert response.json() == {"entities": {"models": 20}}


def test_clear_backend_ok(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.clear = Mock()

    response = test_client.delete("/database", params={"certain": "true"})
    assert response.status_code == 200
    OKMessage.model_validate(response.json())


def test_clear_backend_errors(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.clear = Mock(side_effect=MatchboxDeletionNotConfirmed)

    response = test_client.delete("/database")
    assert response.status_code == 409
    # We send some explanatory message
    assert response.content


def test_api_key_authorisation(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    test_client, _, _ = api_client_and_mocks
    routes = [
        (test_client.post, "/upload/upload_id"),
        (test_client.post, "/resolutions"),
        (test_client.patch, "/resolutions/name/truth"),
        (test_client.delete, "/resolutions/name"),
        (test_client.delete, "/database"),
    ]

    # Incorrect signature
    _, _, signature_b64 = test_client.headers["Authorization"].encode().split(b".")
    header_b64, payload_64, _ = (
        generate_json_web_token(sub="incorrect.user@email.com").encode().split(b".")
    )
    test_client.headers["Authorization"] = b".".join(
        [header_b64, payload_64, signature_b64]
    ).decode()

    for method, url in routes:
        response = method(url)
        assert response.status_code == 401
        assert response.content == b'"JWT invalid."'

    # Expired JWT
    with patch("matchbox.client.authorisation.EXPIRY_AFTER_X_HOURS", -2):
        test_client.headers["Authorization"] = generate_json_web_token(
            sub="test.user@email.com"
        )
        for method, url in routes:
            response = method(url)
            assert response.status_code == 401
            assert response.content == b'"JWT expired."'

    # Missing Authorization header
    test_client.headers.pop("Authorization")
    for method, url in routes:
        response = method(url)
        assert response.status_code == 401
        assert response.content == b'"JWT required but not provided."'


def test_get_resolution_graph(
    resolution_graph: ResolutionGraph,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test the resolution graph report endpoint."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution_graph = Mock(return_value=resolution_graph)

    response = test_client.get("/report/resolutions")
    assert response.status_code == 200
    assert ResolutionGraph.model_validate(response.json())



"""Data transfer objects for Matchbox API."""

import json
import re
import textwrap
from collections.abc import Iterable
from datetime import datetime
from enum import StrEnum
from importlib.metadata import version
from json import JSONDecodeError
from typing import Self

import polars as pl
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from sqlglot import errors, expressions, parse_one

from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_RESULTS
from matchbox.common.graph import (
    ModelResolutionName,
    ResolutionType,
    SourceResolutionName,
)


class DataTypes(StrEnum):
    """Enumeration of supported data types.

    Uses polars datatypes as its backend.
    """

    # Boolean
    BOOLEAN = "Boolean"

    # Integers
    INT8 = "Int8"
    INT16 = "Int16"
    INT32 = "Int32"
    INT64 = "Int64"

    # Unsigned integers
    UINT8 = "UInt8"
    UINT16 = "UInt16"
    UINT32 = "UInt32"
    UINT64 = "UInt64"

    # Floating point
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"

    # Decimal
    DECIMAL = "Decimal"

    # String & Binary
    STRING = "String"
    BINARY = "Binary"

    # Date & Time related
    DATE = "Date"
    TIME = "Time"
    DATETIME = "Datetime"
    DURATION = "Duration"

    # Container types
    ARRAY = "Array"
    LIST = "List"

    # Special types
    OBJECT = "Object"
    CATEGORICAL = "Categorical"
    ENUM = "Enum"
    STRUCT = "Struct"
    NULL = "Null"

    def to_dtype(self) -> pl.DataType:
        """Convert enum value to actual polars dtype."""
        # Map from enum values to actual polars datatypes
        # We do this because polars datatypes are not directly serialisable in Pydantic
        dtype_map = {
            self.BOOLEAN: pl.Boolean,
            self.INT8: pl.Int8,
            self.INT16: pl.Int16,
            self.INT32: pl.Int32,
            self.INT64: pl.Int64,
            self.UINT8: pl.UInt8,
            self.UINT16: pl.UInt16,
            self.UINT32: pl.UInt32,
            self.UINT64: pl.UInt64,
            self.FLOAT32: pl.Float32,
            self.FLOAT64: pl.Float64,
            self.DECIMAL: pl.Decimal,
            self.STRING: pl.String,
            self.BINARY: pl.Binary,
            self.DATE: pl.Date,
            self.TIME: pl.Time,
            self.DATETIME: pl.Datetime,
            self.DURATION: pl.Duration,
            self.ARRAY: pl.Array,
            self.LIST: pl.List,
            self.OBJECT: pl.Object,
            self.CATEGORICAL: pl.Categorical,
            self.ENUM: pl.Enum,
            self.STRUCT: pl.Struct,
            self.NULL: pl.Null,
        }
        return dtype_map[self]

    def to_pytype(self) -> type:
        """Convert enum value to actual Python type."""
        return self.to_dtype().to_python()

    @classmethod
    def from_dtype(cls, dtype: pl.DataType) -> "DataTypes":
        """Get enum value from a polars dtype."""
        # Find the name of the dtype class
        dtype_name = dtype.__class__.__name__
        # Find the matching enum value
        for enum_val in cls:
            if enum_val.value == dtype_name:
                return enum_val
        raise ValueError(f"No matching polars DataTypes for dtype: {dtype_name}")

    @classmethod
    def from_pytype(cls, pytype: type) -> "DataTypes":
        """Get enum value from a Python type."""
        return DataTypes.from_dtype(pl.DataType.from_python(pytype))


class OKMessage(BaseModel):
    """Generic HTTP OK response."""

    status: str = Field(default="OK")
    version: str = Field(default_factory=lambda: version("matchbox-db"))


class LoginAttempt(BaseModel):
    """Request for log in process."""

    user_name: str


class LoginResult(BaseModel):
    """Response from log in process."""

    user_id: int


class BackendCountableType(StrEnum):
    """Enumeration of supported backend countable types."""

    SOURCES = "sources"
    MODELS = "models"
    DATA = "data"
    CLUSTERS = "clusters"
    CREATES = "creates"
    MERGES = "merges"
    PROPOSES = "proposes"


class ModelResultsType(StrEnum):
    """Enumeration of supported model results types."""

    PROBABILITIES = "probabilities"
    CLUSTERS = "clusters"


class BackendResourceType(StrEnum):
    """Enumeration of resources types referenced by client or API."""

    RESOLUTION = "resolution"
    CLUSTER = "cluster"
    USER = "user"
    JUDGEMENT = "judgement"


class BackendParameterType(StrEnum):
    """Enumeration of parameters passable to the API."""

    SAMPLE_SIZE = "sample_size"


class BackendUploadType(StrEnum):
    """Enumeration of supported backend upload types."""

    INDEX = "index"
    RESULTS = "results"

    @property
    def schema(self):
        """Get the schema for the upload type."""
        return {
            BackendUploadType.INDEX: SCHEMA_INDEX,
            BackendUploadType.RESULTS: SCHEMA_RESULTS,
        }[self]


class CRUDOperation(StrEnum):
    """Enumeration of CRUD operations."""

    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


class LocationType(StrEnum):
    """Enumeration of location types."""

    RDBMS = "rdbms"


class LocationConfig(BaseModel):
    """Metadata for a location."""

    model_config = ConfigDict(frozen=True)

    type: LocationType
    name: str


class SourceField(BaseModel):
    """A field in a source that can be indexed in the Matchbox database."""

    model_config = ConfigDict(frozen=True)

    name: str = Field(
        description=(
            "The name of the field in the source after the "
            "extract/transform logic has been applied."
        )
    )
    type: DataTypes = Field(
        description="The cached field type. Used to ensure a stable hash.",
    )


class SourceConfig(BaseModel):
    """Configuration of a source that can, or has been, indexed in the backend.

    They are foundational processes on top of which linking and deduplication models can
    build new resolutions.
    """

    model_config = ConfigDict(frozen=True)

    location_config: LocationConfig = Field(
        description=(
            "The location of the source. Used to run the extract/tansform logic."
        ),
    )
    extract_transform: str = Field(
        description=(
            "Logic to extract and transform data from the source. "
            "Language is location dependent."
        )
    )
    # Fields can to be set at creation, or initialised with `.default_columns()`
    key_field: SourceField = Field(
        description=textwrap.dedent("""
            The key field. This is the source's key for unique
            entities, such as a primary key in a relational database.

            Keys must ALWAYS be a string.

            For example, if the source describes companies, it may have used
            a Companies House number as its key.

            This key is ALWAYS correct. It should be something generated and
            owned by the source being indexed.
            
            For example, your organisation's CRM ID is a key field within the CRM.
            
            A CRM ID entered by hand in another dataset shouldn't be used 
            as a key field.
        """),
    )
    index_fields: tuple[SourceField, ...] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The fields to index in this source, after the extract/transform logic 
            has been applied. 

            This is usually set manually, and should map onto the columns that the
            extract/transform logic returns.
            """
        ),
    )

    @model_validator(mode="after")
    def validate_key_field(self) -> Self:
        """Ensure that the key field is a string and not in the index fields."""
        if self.key_field in self.index_fields:
            raise ValueError("Key field must not be in the index fields. ")

        if self.key_field.type != DataTypes.STRING:
            raise ValueError("Key field must be a string. ")

        return self

    def prefix(self, name: str) -> str:
        """Get the prefix for the source.

        Args:
            name: The name of the source.

        Returns:
            The prefix string (name + "_").
        """
        return name + "_"

    def qualified_key(self, name: str) -> str:
        """Get the qualified key for the source.

        Args:
            name: The name of the source.

        Returns:
            The qualified key field name.
        """
        return self.qualify_field(name, self.key_field.name)

    def qualified_index_fields(self, name: str) -> list[str]:
        """Get the qualified index fields for the source.

        Args:
            name: The name of the source.

        Returns:
            List of qualified index field names.
        """
        return [self.qualify_field(name, field.name) for field in self.index_fields]

    def qualify_field(self, name: str, field: str) -> str:
        """Qualify field names with the source name.

        Args:
            name: The name of the source.
            field: The field name to qualify.

        Returns:
            A single qualified field.
        """
        return self.prefix(name) + field

    def f(self, name: str, fields: str | Iterable[str]) -> str | list[str]:
        """Qualify one or more field names with the source name.

        Args:
            name: The name of the source.
            fields: The field name to qualify, or a list of field names.

        Returns:
            A single qualified field, or a list of qualified field names.
        """
        if isinstance(fields, str):
            return self.qualify_field(name, fields)
        return [self.qualify_field(name, field_name) for field_name in fields]


class QueryCombineType(StrEnum):
    """Enumeration of ways to combine multiple rows having the same matchbox ID."""

    CONCAT = "concat"
    EXPLODE = "explode"
    SET_AGG = "set_agg"


class QueryConfig(BaseModel):
    """Configuration of query generating model inputs."""

    model_config = ConfigDict(frozen=True)

    source_resolutions: tuple[SourceResolutionName, ...]
    model_resolution: ModelResolutionName | None = None
    combine_type: QueryCombineType = QueryCombineType.CONCAT
    threshold: int | None = None
    cleaning: dict[str, str] | None = None

    @model_validator(mode="after")
    def validate_resolutions(self) -> Self:
        """Ensure that resolution settings are compatible."""
        if not self.source_resolutions:
            raise ValueError("At least one source resolution required.")
        if len(self.source_resolutions) > 1 and not self.model_resolution:
            raise ValueError(
                "A model resolution must be set if querying from multiple sources"
            )
        return self

    @field_validator("cleaning")
    @classmethod
    def validate_cleaning_dict(cls, v: dict[str, str] | None) -> str | None:
        """Validate cleaning as valid SQL."""
        if v is None:
            return v

        for alias, sql in v.items():
            if sql is not None:
                try:
                    stmt = parse_one(sql, dialect="duckdb")
                except errors.ParseError as e:
                    raise ValueError(f"Invalid SQL in cleaning_dict: {alias}") from e

                for node in stmt.walk():
                    if isinstance(node, expressions.Column) and node.name == "id":
                        raise ValueError(
                            "Cannot transform 'id' column in cleaning_dict. "
                            "It is always selected by default."
                        )

        return v

    @property
    def dependencies(self) -> list[str]:
        """Return all resolution names that this query needs."""
        deps = list(self.source_resolutions)
        if self.model_resolution:
            deps.append(self.model_resolution)

        return deps

    @property
    def point_of_truth(self):
        """Return name of resolution that will be used as point of truth."""
        if self.model_resolution:
            return self.model_resolution
        return self.source_resolutions[0]


class ModelType(StrEnum):
    """Enumeration of supported model types."""

    LINKER = "linker"
    DEDUPER = "deduper"


class ModelConfig(BaseModel):
    """Configuration for model that has or could be added to the server."""

    type: ModelType
    model_class: str
    model_settings: str
    left_query: QueryConfig
    right_query: QueryConfig | None = None  # Only used for linker models

    def __eq__(self, other: Self) -> bool:
        """Check equality of model configurations.

        Model configurations don't care about the order of left and right resolutions.
        """
        if not isinstance(other, ModelConfig):
            return NotImplemented
        return self.type == other.type and {
            self.left_query,
            self.right_query,
        } == {other.left_query, other.right_query}

    @model_validator(mode="after")
    def validate_right_query(self) -> Self:
        """Ensure that a right query is set if and only if model is linker."""
        if self.type == ModelType.DEDUPER and self.right_query is not None:
            raise ValueError("Right query can't be set for dedupers")
        if self.type == ModelType.LINKER and self.right_query is None:
            raise ValueError("Right query must be set for linkers")
        return self

    @field_validator("model_settings", mode="after")
    @classmethod
    def validate_settings_json(cls, value: str) -> str:
        """Ensure that the model settings is valid JSON."""
        try:
            isinstance(json.loads(value), dict)
        except JSONDecodeError as e:
            raise ValueError("Model settings are not valid JSON") from e
        return value


class Match(BaseModel):
    """A match between primary keys in the Matchbox database."""

    cluster: int | None
    source: str
    source_id: set[str] = Field(default_factory=set)
    target: str
    target_id: set[str] = Field(default_factory=set)

    @model_validator(mode="after")
    def found_or_none(self) -> "Match":
        """Ensure that a match has sources and a cluster if target was found."""
        if self.target_id and not (self.source_id and self.cluster):
            raise ValueError(
                "A match must have sources and a cluster if target was found."
            )
        if self.cluster and not self.source_id:
            raise ValueError("A match must have source if cluster is set.")
        return self

    @field_serializer("source_id", "target_id")
    def serialise_ids(self, id_set: set[str]):
        """Turn set to sorted list when serialising."""
        return sorted(id_set)


class ModelAncestor(BaseModel):
    """A model's ancestor and its truth value."""

    name: ModelResolutionName = Field(..., description="Name of the ancestor model")
    truth: int | None = Field(
        default=None, description="Truth threshold value", ge=0, le=100, strict=True
    )


class Resolution(BaseModel):
    """Unified resolution type with common fields and discriminated config."""

    name: str = Field(description="Unique name of the resolution")
    description: str | None = Field(default=None, description="Description")
    truth: int | None = Field(default=None, ge=0, le=100, strict=True)

    # Discriminator field
    resolution_type: ResolutionType

    # Type-specific config as discriminated union
    config: SourceConfig | ModelConfig

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, value: str) -> str:
        """Ensure the name is a valid resolution name.

        Raises:
            ValueError: If the name is not a valid resolution name.
        """
        if not re.match(r"^[a-zA-Z0-9_]+$", value):
            raise ValueError(
                "Resolution names must be alphanumeric and underscore only."
            )
        return value

    @field_validator("description", mode="after")
    @classmethod
    def validate_description(cls, value: str | None) -> str | None:
        """Ensure the description is not empty if provided."""
        if value is not None and not value.strip():
            raise ValueError("Description cannot be empty if provided.")
        return value

    @model_validator(mode="after")
    def validate_resolution_type_matches_config(self):
        """Ensure resolution_type matches the config type."""
        if self.resolution_type == ResolutionType.SOURCE:
            assert isinstance(self.config, SourceConfig), (
                "Config must be SourceConfig when resolution_type is 'source'"
            )
        else:
            assert isinstance(self.config, ModelConfig), (
                "Config must be ModelConfig when resolution_type is 'model'"
            )
        return self

    @model_validator(mode="after")
    def validate_truth_matches_type(self):
        """Ensure truth field matches resolution type requirements."""
        if self.resolution_type == ResolutionType.SOURCE and self.truth is not None:
            raise ValueError("Truth must be None for source resolutions")
        elif self.resolution_type == ResolutionType.MODEL:
            if self.truth is None:
                raise ValueError("Truth is required for model resolutions")
            if not (0 <= self.truth <= 100):
                raise ValueError(
                    "Truth must be between 0 and 100 for model resolutions"
                )
        return self


class ResolutionOperationStatus(BaseModel):
    """Status response for any resolution operation."""

    success: bool
    name: ModelResolutionName
    operation: CRUDOperation
    details: str | None = None

    @classmethod
    def status_409_examples(cls) -> dict:
        """Examples for 409 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "confirm_delete": {
                            "summary": "Delete operation requires confirmation. ",
                            "value": cls(
                                success=False,
                                name="example_model",
                                operation=CRUDOperation.DELETE,
                                details=(
                                    "This operation will delete the resolutions "
                                    "deduper_1, deduper_2, linker_1, "
                                    "as well as all probabilities they have created. "
                                    "\n\n"
                                    "It won't delete validation associated with these "
                                    "probabilities. \n\n"
                                    "If you're sure you want to continue, rerun with "
                                    "certain=True"
                                ),
                            ).model_dump(),
                        },
                    },
                }
            }
        }

    @classmethod
    def status_500_examples(cls) -> dict:
        """Examples for 500 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "unhandled": {
                            "summary": (
                                "Unhandled exception encountered while updating the "
                                "model's truth value."
                            ),
                            "value": cls(
                                success=False,
                                name="example_model",
                                operation=CRUDOperation.UPDATE,
                            ).model_dump(),
                        },
                    },
                }
            }
        }


class CountResult(BaseModel):
    """Response model for count results."""

    entities: dict[BackendCountableType, int]


class UploadStage(StrEnum):
    """Enumeration of stages of a file upload and its processing."""

    READY = "ready"
    AWAITING_UPLOAD = "awaiting_upload"
    QUEUED = "queued"
    PROCESSING = "processing"
    COMPLETE = "complete"
    FAILED = "failed"
    UNKNOWN = "unknown"


class UploadStatus(BaseModel):
    """Response model for any file upload processes."""

    id: str
    stage: UploadStage
    update_timestamp: datetime
    details: str | None = None
    entity: BackendUploadType | None = None

    _status_code_mapping = {
        UploadStage.READY: 200,
        UploadStage.COMPLETE: 200,
        UploadStage.FAILED: 400,
        UploadStage.AWAITING_UPLOAD: 202,
        UploadStage.QUEUED: 200,
        UploadStage.PROCESSING: 200,
    }

    def get_http_code(self) -> int:
        """Get the HTTP status code for the upload stage."""
        if self.stage == UploadStage.FAILED:
            return 400
        return self._status_code_mapping[self.stage]

    @classmethod
    def status_400_examples(cls) -> dict:
        """Examples for 400 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "expired_id": {
                            "summary": "Upload ID expired",
                            "value": cls(
                                id="example_id",
                                stage=UploadStage.FAILED,
                                details=(
                                    "Upload ID not found or expired. Entries expire "
                                    "after 30 minutes of inactivity, including "
                                    "failed processes."
                                ),
                                entity=BackendUploadType.INDEX,
                                update_timestamp=datetime.now(),
                            ).model_dump(),
                        },
                        "schema_mismatch": {
                            "summary": "Schema validation error",
                            "value": cls(
                                id="example_id",
                                stage=UploadStage.FAILED,
                                details="Schema mismatch. Expected: ... Got: ...",
                                entity=BackendUploadType.INDEX,
                                update_timestamp=datetime.now(),
                            ).model_dump(),
                        },
                    },
                }
            }
        }


class NotFoundError(BaseModel):
    """API error for a 404 status code."""

    details: str
    entity: BackendResourceType


class InvalidParameterError(BaseModel):
    """API error for a custom 422 status code."""

    details: str
    parameter: BackendParameterType



"""Worker logic to process user uploads."""

import uuid
from abc import ABC, abstractmethod
from collections.abc import Generator
from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING, Any

import pyarrow as pa
import redis
from celery import Celery, Task
from fastapi import UploadFile
from pyarrow import parquet as pq
from pydantic import BaseModel

from matchbox.common.dtos import (
    BackendUploadType,
    Resolution,
    UploadStage,
    UploadStatus,
)
from matchbox.common.exceptions import MatchboxServerFileError
from matchbox.common.graph import ResolutionType
from matchbox.common.logging import logger
from matchbox.server.base import (
    MatchboxDBAdapter,
    MatchboxServerSettings,
    get_backend_settings,
    settings_to_backend,
)

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any

from celery.exceptions import MaxRetriesExceededError
from celery.utils.log import get_task_logger

celery_logger = get_task_logger(__name__)

# -- Upload trackers --


class UploadEntry(BaseModel):
    """Entry in upload tracker, combining private metadata and public upload status."""

    status: UploadStatus
    metadata: Resolution


class UploadTracker(ABC):
    """Abstract class for upload tracker."""

    @staticmethod
    def _create_entry(
        metadata: Resolution, upload_type: BackendUploadType
    ) -> UploadEntry:
        """Create initial UploadEntry object."""
        upload_id = str(uuid.uuid4())

        return UploadEntry(
            metadata=metadata,
            status=UploadStatus(
                id=upload_id,
                stage=UploadStage.AWAITING_UPLOAD,
                update_timestamp=datetime.now(),
                entity=upload_type,
            ),
        )

    def _get_updated_entry(
        self, upload_id: str, stage: str, details: str | None
    ) -> UploadEntry:
        """Create new UploadEntry object as update on previous entry."""
        entry = self.get(upload_id)
        if not entry:
            raise KeyError(f"Entry {upload_id} not found.")

        status = entry.status.model_copy(
            update={"stage": stage, "update_timestamp": datetime.now()}
        )
        if details:
            status.details = details

        return UploadEntry(status=status, metadata=entry.metadata)

    def add_source(self, metadata: Resolution) -> str:
        """Register source resolution and return ID."""
        assert metadata.resolution_type == ResolutionType.SOURCE
        entry = self._create_entry(metadata, BackendUploadType.INDEX)
        self._register_entry(entry)

        return entry.status.id

    def add_model(self, metadata: Resolution) -> str:
        """Register model resolution and return ID."""
        assert metadata.resolution_type == ResolutionType.MODEL
        entry = self._create_entry(metadata, BackendUploadType.RESULTS)
        self._register_entry(entry)

        return entry.status.id

    def add_resolution(self, metadata: Resolution) -> str:
        """Generic add that routes based on type."""
        if metadata.resolution_type == ResolutionType.SOURCE:
            return self.add_source(metadata)
        else:
            return self.add_model(metadata)

    @abstractmethod
    def _register_entry(self, UploadEntry) -> str:
        """Register UploadEntry object to tracker and return its ID."""
        ...

    @abstractmethod
    def get(self, upload_id: str) -> UploadEntry | None:
        """Retrieve metadata by ID if not expired."""
        ...

    @abstractmethod
    def update(self, upload_id: str, stage: str, details: str | None = None) -> None:
        """Update the stage and details for an upload.

        Raises:
            KeyError: If entry not found.
        """
        ...


class InMemoryUploadTracker(UploadTracker):
    """In-memory upload tracker, only usable with single server instance."""

    def __init__(self):
        """Initialise tracker data structure."""
        self._tracker = {}

    def _register_entry(self, entry: UploadEntry) -> None:
        self._tracker[entry.status.id] = entry

    def get(self, upload_id: str) -> UploadEntry | None:  # noqa: D102
        return self._tracker.get(upload_id)

    def update(  # noqa: D102
        self, upload_id: str, stage: str, details: str | None = None
    ) -> None:
        self._tracker[upload_id] = self._get_updated_entry(
            upload_id=upload_id, stage=stage, details=details
        )


class RedisUploadTracker(UploadTracker):
    """Upload tracker backed by Redis."""

    def __init__(self, redis_url: str, expiry_minutes: int, key_space: str = "upload"):
        """Connect Redis and initialise tracker object."""
        self.expiry_minutes = expiry_minutes
        self.redis = redis.Redis.from_url(redis_url)
        self.key_prefix = f"{key_space}:"

    def _to_redis(self, key: str, value: str):
        expiry_seconds = self.expiry_minutes * 60
        self.redis.setex(f"{self.key_prefix}{key}", expiry_seconds, value)

    def _register_entry(self, entry: UploadEntry) -> str:  # noqa: D102
        self._to_redis(entry.status.id, entry.model_dump_json())

        return entry.status.id

    def get(self, upload_id: str) -> UploadEntry | None:  # noqa: D102
        data = self.redis.get(f"{self.key_prefix}{upload_id}")
        if not data:
            return None

        entry = UploadEntry.model_validate_json(data)

        return entry

    def update(  # noqa: D102
        self, upload_id: str, stage: str, details: str | None = None
    ) -> None:
        entry = self._get_updated_entry(
            upload_id=upload_id, stage=stage, details=details
        )

        self._to_redis(upload_id, entry.model_dump_json())


_IN_MEMORY_TRACKER = InMemoryUploadTracker()


def settings_to_upload_tracker(settings: MatchboxServerSettings) -> UploadTracker:
    """Initialise an upload tracker from server settings."""
    match settings.task_runner:
        case "api":
            return _IN_MEMORY_TRACKER
        case "celery":
            return RedisUploadTracker(
                redis_url=settings.redis_uri,
                expiry_minutes=settings.uploads_expiry_minutes,
            )
        case _:
            raise RuntimeError("Unsupported task runner.")


# -- S3 functions --


def table_to_s3(
    client: S3Client,
    bucket: str,
    key: str,
    file: UploadFile,
    expected_schema: pa.Schema,
) -> str:
    """Upload a PyArrow Table to S3 and return the key.

    Args:
        client: The S3 client to use.
        bucket: The S3 bucket to upload to.
        key: The key to upload to.
        file: The file to upload.
        expected_schema: The schema that the file should match.

    Raises:
        MatchboxServerFileError: If the file is not a valid Parquet file or the schema
            does not match the expected schema.

    Returns:
        The key of the uploaded file.
    """
    try:
        table = pq.read_table(file.file)

        if not table.schema.equals(expected_schema):
            raise MatchboxServerFileError(
                message=(
                    "Schema mismatch. "
                    f"Expected:\n{expected_schema}\nGot:\n{table.schema}"
                )
            )

        file.file.seek(0)

        client.put_object(Bucket=bucket, Key=key, Body=file.file)

    except Exception as e:
        if isinstance(e, MatchboxServerFileError):
            raise
        raise MatchboxServerFileError(message=f"Invalid Parquet file: {str(e)}") from e

    return key


def s3_to_recordbatch(
    client: S3Client, bucket: str, key: str, batch_size: int = 1000
) -> Generator[pa.RecordBatch, None, None]:
    """Download a PyArrow Table from S3 and stream it as RecordBatches."""
    response = client.get_object(Bucket=bucket, Key=key)
    buffer = pa.BufferReader(response["Body"].read())

    parquet_file = pq.ParquetFile(buffer)

    yield from parquet_file.iter_batches(batch_size=batch_size)


# -- Upload tasks --


CELERY_SETTINGS = get_backend_settings(MatchboxServerSettings().backend_type)()
CELERY_BACKEND: MatchboxDBAdapter | None = None
CELERY_TRACKER: UploadTracker | None = None

celery = Celery("matchbox", broker=CELERY_SETTINGS.redis_uri)
celery.conf.update(
    # Hard time limit for tasks (in seconds)
    task_time_limit=CELERY_SETTINGS.uploads_expiry_minutes * 60,
    # Only acknowledge task (remove it from queue) after task completion
    task_acks_late=True,
    # Reduce pre-fetching (workers reserving tasks while they're still busy)
    # as it's not ideal for long-running tasks
    prefetch_multiplier=1,
)


def initialise_celery_worker():
    """Initialise backend and tracker for celery worker."""
    global CELERY_SETTINGS
    global CELERY_BACKEND
    global CELERY_TRACKER

    if not CELERY_BACKEND:
        CELERY_BACKEND = settings_to_backend(CELERY_SETTINGS)
    if not CELERY_TRACKER:
        CELERY_TRACKER = settings_to_upload_tracker(CELERY_SETTINGS)


def process_upload(
    backend: MatchboxDBAdapter,
    tracker: UploadTracker,
    s3_client: S3Client,
    upload_type: str,
    resolution_name: str,
    upload_id: str,
    bucket: str,
    filename: str,
) -> None:
    """Generic task to process uploaded file, usable by FastAPI BackgroundTasks."""
    tracker.update(upload_id, UploadStage.PROCESSING)
    upload = tracker.get(upload_id)

    try:
        data = pa.Table.from_batches(
            [
                batch
                for batch in s3_to_recordbatch(
                    client=s3_client, bucket=bucket, key=filename
                )
            ]
        )

        if upload.status.entity == BackendUploadType.INDEX:
            backend.insert_source_data(name=upload.metadata.name, data_hashes=data)
        elif upload.status.entity == BackendUploadType.RESULTS:
            backend.insert_model_data(name=upload.metadata.name, results=data)
        else:
            raise ValueError(f"Unknown upload type: {upload.status.entity}")

        tracker.update(upload_id, UploadStage.COMPLETE)

    except Exception as e:
        error_context = {
            "upload_id": upload_id,
            "upload_type": upload_type,
            "resolution_name": resolution_name,
            "bucket": bucket,
            "key": filename,
        }
        logger.error(
            f"Upload processing failed with context: {error_context}", exc_info=True
        )
        details = (
            f"Error: {e}. Context: "
            f"Upload type: {getattr(upload.status, 'entity', 'unknown')}, "
            f"SourceConfig: {getattr(upload, 'metadata', 'unknown')}"
        )
        tracker.update(
            upload_id,
            UploadStage.FAILED,
            details=details,
        )
        raise MatchboxServerFileError(message=details) from e
    finally:
        try:
            s3_client.delete_object(Bucket=bucket, Key=filename)
        except Exception as delete_error:  # noqa: BLE001
            logger.error(
                f"Failed to delete S3 file {bucket}/{filename}: {delete_error}"
            )


@celery.task(ignore_result=True, bind=True, max_retries=3)
def process_upload_celery(
    self: Task,
    upload_type: str,
    resolution_name: str,
    upload_id: str,
    bucket: str,
    filename: str,
) -> None:
    """Celery task to process uploaded file, with only serialisable arguments."""
    initialise_celery_worker()

    celery_logger.info(
        "Uploading data for resolution %s, ID %s", resolution_name, upload_id
    )

    upload_function = partial(
        process_upload,
        backend=CELERY_BACKEND,
        tracker=CELERY_TRACKER,
        s3_client=CELERY_BACKEND.settings.datastore.get_client(),
    )

    try:
        upload_function(
            upload_type=upload_type,
            resolution_name=resolution_name,
            upload_id=upload_id,
            bucket=bucket,
            filename=filename,
        )
    except Exception as exc:  # noqa: BLE001
        celery_logger.error(
            "Upload failed for resolution %s, ID %s. Retrying...",
            resolution_name,
            upload_id,
        )
        try:
            raise self.retry(exc=exc) from None
        except MaxRetriesExceededError:
            if CELERY_TRACKER:
                CELERY_TRACKER.update(
                    upload_id, UploadStage.FAILED, f"Max retries exceeded: {exc}"
                )
            raise

    celery_logger.info("Upload complete for %s, ID %s", resolution_name, upload_id)



"""Functions to transform data between tabular and graph structures."""

from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable
from typing import Any, Generic, Self, TypeVar

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import rustworkx as rx

from matchbox.common.hash import HASH_FUNC, IntMap, hash_values

T = TypeVar("T", bound=Hashable)


def to_clusters(
    results: pa.Table,
    dtype: pa.DataType = pa.large_binary,
    hash_func: Callable[[*tuple[T, ...]], T] = hash_values,
) -> pa.Table:
    """Converts probabilities into connected components formed at each threshold.

    Args:
        results: Arrow table with columns ['left_id', 'right_id', 'probability']
        dtype: Arrow data type for the parent and child columns
        hash_func: Function to hash the parent and child values

    Returns:
        Arrow table of parent, child, threshold, sorted by probability descending.
    """
    G = rx.PyGraph()
    added: dict[bytes, int] = {}
    components: dict[str, list] = {"parent": [], "child": [], "threshold": []}

    # Sort probabilities descending and select relevant columns
    results = results.sort_by([("probability", "descending")])
    results = results.select(["left_id", "right_id", "probability"])

    # Get unique probability thresholds, sorted
    thresholds = pc.unique(results.column("probability")).sort(order="descending")

    # Process edges grouped by probability threshold
    for prob in thresholds:
        threshold_edges = results.filter(pc.equal(results.column("probability"), prob))

        # Get state before adding this batch of edges
        old_components = {frozenset(comp) for comp in rx.connected_components(G)}

        # Add all nodes and edges at this probability threshold
        edge_values = list(
            zip(
                threshold_edges.column("left_id").to_pylist(),
                threshold_edges.column("right_id").to_pylist(),
                strict=True,
            )
        )
        for left, right in edge_values:
            for hash_val in (left, right):
                if hash_val not in added:
                    idx = G.add_node(hash_val)
                    added[hash_val] = idx

            G.add_edge(added[left], added[right], None)

        new_components = {frozenset(comp) for comp in rx.connected_components(G)}
        changed_components = new_components - old_components

        # For each changed component, add ALL members at current threshold
        for comp in changed_components:
            children = [G.get_node_data(n) for n in comp]
            parent = hash_func(*children)

            components["parent"].extend([parent] * len(children))
            components["child"].extend(children)
            components["threshold"].extend([prob] * len(children))

    return pa.Table.from_pydict(
        components,
        schema=pa.schema(
            [("parent", dtype()), ("child", dtype()), ("threshold", pa.uint8())]
        ),
    )


def graph_results(
    probabilities: pa.Table, all_nodes: Iterable[int] | None = None
) -> tuple[rx.PyDiGraph, np.ndarray, np.ndarray]:
    """Convert probability table to graph representation.

    Args:
        probabilities: PyArrow table with 'left_id', 'right_id' columns
        all_nodes: superset of node identities figuring in probabilities table.
            Used to optionally add isolated nodes to the graph.

    Returns:
        A tuple containing:
        - Rustwork directed graph
        - A list mapping the 'left_id' probabilities to node indices in the graph
        - A list mapping the 'right_id' probabilities to node indices in the graph
    """
    # Create index to use in graph
    unique = pc.unique(
        pa.concat_arrays(
            [
                probabilities["left_id"].combine_chunks(),
                probabilities["right_id"].combine_chunks(),
            ]
        )
    )

    left_indices = pc.index_in(probabilities["left_id"], unique).to_numpy()
    right_indices = pc.index_in(probabilities["right_id"], unique).to_numpy()

    # Create and process graph
    n_nodes = len(unique)
    n_edges = len(probabilities)

    graph = rx.PyGraph(node_count_hint=n_nodes, edge_count_hint=n_edges)
    graph.add_nodes_from(range(n_nodes))

    if all_nodes is not None:
        isolated_nodes = len(set(all_nodes) - set(unique.to_pylist()))
        graph.add_nodes_from(range(isolated_nodes))

    edges = tuple(zip(left_indices, right_indices, strict=True))
    graph.add_edges_from_no_data(edges)

    return graph, left_indices, right_indices


class DisjointSet(Generic[T]):
    """Disjoint set forest with "path compression" and "union by rank" heuristics.

    This follows implementation from Cormen, Thomas H., et al. Introduction to
    algorithms. MIT press, 2022
    """

    def __init__(self):
        """Initialize the disjoint set."""
        self.parent: dict[T, T] = {}
        self.rank: dict[T, int] = {}

    def _make_set(self, x: T) -> None:
        """Create a new set with a single element x."""
        self.parent[x] = x
        self.rank[x] = 0

    def add(self, x: T) -> None:
        """Add a new element to the disjoint set."""
        if x not in self.parent:
            self._make_set(x)

    def union(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        self._link(self._find(x), self._find(y))

    def _link(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        if self.rank[x] > self.rank[y]:
            self.parent[y] = x
        else:
            self.parent[x] = y
            if self.rank[x] == self.rank[y]:
                self.rank[y] += 1

    def _find(self, x: T) -> T:
        """Return the representative element of the set containing x."""
        if x not in self.parent:
            self._make_set(x)
            return x

        if x != self.parent[x]:
            self.parent[x] = self._find(self.parent[x])

        return self.parent[x]

    def get_components(self) -> list[set[T]]:
        """Return the connected components of the disjoint set."""
        components = defaultdict(set)
        for x in self.parent:
            root = self._find(x)
            components[root].add(x)
        return list(components.values())


def hash_cluster_leaves(leaves: list[bytes]) -> bytes:
    """Canonical method to convert list of cluster IDs to their combined hash."""
    return HASH_FUNC(b"|".join(leaf for leaf in sorted(leaves))).digest()


class Cluster:
    """A cluster of connected components.

    Can be a source cluster (a single data point) or a model cluster (a cluster of
    clusters). The hash of a cluster is the hash of its source clusters -- its leaves.

    We generate negative integers for IDs, allowing us to generate a true ID with
    the database after we've calculated components using these objects.
    """

    id: int
    probability: int | None
    hash: bytes
    leaves: tuple["Cluster"] | None

    _intmap: IntMap  # Reference to the IntMap singleton

    def __init__(
        self,
        intmap: IntMap,
        probability: int | None = None,
        leaves: tuple["Cluster"] | list["Cluster"] | None = None,
        id: int | None = None,
        hash: bytes | None = None,
    ):
        """Initialise the Cluster.

        Args:
            intmap: An IntMap instance for generating unique IDs
            probability: probability of the cluster from its resolution, or None if
                source
            leaves: A list of Cluster objects that are the leaves of this cluster
            id: The ID of the cluster (only for leaf nodes)
            hash: The hash of the cluster (only for leaf nodes)
        """
        self._intmap = intmap
        self.probability = probability

        if leaves:
            self.leaves = tuple(sorted(leaves, key=lambda leaf: leaf.hash))
        else:
            self.leaves = None

        # Set hash - use provided hash or calculate
        if hash is not None:
            self.hash = hash
        elif self.leaves is None:
            raise ValueError("Leaf nodes must have hash specified")
        else:
            self.hash = hash_cluster_leaves([leaf.hash for leaf in self.leaves])

        # Set ID - use provided ID or calculate
        if id is not None:
            self.id = id
        elif self.leaves is None:
            raise ValueError("Leaf nodes must have id specified")
        else:
            self.id = intmap.index(leaf.id for leaf in self.leaves)

    def __hash__(self) -> int:
        """Return a hash of the cluster based on its hash bytes."""
        return hash(self.hash)

    def __eq__(self, other: Any) -> bool:
        """Check equality of two Cluster objects based on their hash."""
        if not isinstance(other, Cluster):
            return False
        return self.hash == other.hash

    @classmethod
    def combine(
        cls: type[Self],
        clusters: Iterable["Cluster"],
        probability: int | None = None,
    ) -> "Cluster":
        """Efficiently combine multiple clusters at once.

        Args:
            clusters: An iterable of Cluster objects to combine
            probability: the probability of the cluster from its resolution

        Returns:
            A new Cluster containing all unique leaves from the input clusters
        """
        clusters = list(clusters)
        if len(clusters) == 1:
            new_probability = (
                clusters[0].probability if probability is None else probability
            )
            return cls(
                intmap=clusters[0]._intmap,
                probability=new_probability,
                leaves=clusters[0].leaves,
                id=clusters[0].id,
                hash=clusters[0].hash,
            )

        intmap = clusters[0]._intmap
        unique_dict: dict[int, Cluster] = {}

        for cluster in clusters:
            if cluster.leaves is None:
                unique_dict[id(cluster)] = cluster
            else:
                for leaf in cluster.leaves:
                    unique_dict[id(leaf)] = leaf

        return cls(
            intmap=intmap,
            probability=probability,
            leaves=list(unique_dict.values()),
        )



"""Functions and classes to define, run and register models."""

import inspect
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from matchbox.client import _handler
from matchbox.client._settings import settings
from matchbox.client.models import dedupers, linkers
from matchbox.client.models.dedupers.base import Deduper, DeduperSettings
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.client.results import Results
from matchbox.common.dtos import ModelConfig, ModelType, Resolution
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.graph import ResolutionType
from matchbox.common.logging import logger

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.queries import Query
else:
    DAG = Any
    Query = Any

P = ParamSpec("P")
R = TypeVar("R")


_MODEL_CLASSES = {
    **{name: obj for name, obj in inspect.getmembers(dedupers, inspect.isclass)},
    **{name: obj for name, obj in inspect.getmembers(linkers, inspect.isclass)},
}


def add_model_class(ModelClass: type[Linker] | type[Deduper]) -> None:
    """Add custom deduper or linker."""
    if issubclass(ModelClass, Linker) or issubclass(ModelClass, Deduper):
        _MODEL_CLASSES[ModelClass.__name__] = ModelClass
    else:
        raise ValueError("The argument is not a proper subclass of Deduper or Linker.")


class Model:
    """Unified model class for both linking and deduping operations."""

    @overload
    def __init__(
        self,
        name: str,
        dag: DAG,
        model_class: type[Deduper],
        model_settings: DeduperSettings | dict,
        left_query: Query,
        right_query: None = None,
        truth: float = 1.0,
        description: str | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Linker],
        model_settings: LinkerSettings | dict,
        left_query: Query,
        right_query: Query,
        truth: float = 1.0,
        description: str | None = None,
    ) -> None: ...

    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Deduper] | type[Linker] | str,
        model_settings: DeduperSettings | LinkerSettings | dict,
        left_query: Query,
        right_query: Query | None = None,
        truth: float = 1.0,
        description: str | None = None,
    ):
        """Create a new model instance.

        Args:
            dag: DAG containing this model.
            name: Unique name for the model
            truth: Truth threshold. Defaults to 1.0. Can be set later after analysis.
            model_class: Class of Linker or Deduper, or its name.
            model_settings: Appropriate settings object to pass to model class.
            left_query: The query that will get the data to deduplicate, or the data to
                link on the left.
            right_query: The query that will get the data to link on the right.
            description: Optional description of the model
        """
        self.last_run: datetime | None = None
        self.dag = dag
        self.name = name
        self.description = description
        self._truth: int = _truth_float_to_int(truth)
        self.left_query = left_query
        self.right_query = right_query
        self.results: Results | None = None

        if isinstance(model_class, str):
            model_class: type[Linker | Deduper] = _MODEL_CLASSES[model_class]
        self.model_instance = model_class(settings=model_settings)

        model_type: ModelType = (
            ModelType.LINKER if issubclass(model_class, Linker) else ModelType.DEDUPER
        )

        if isinstance(model_settings, dict):
            SettingsClass = self.model_instance.__annotations__["settings"]
            model_settings = SettingsClass(**model_settings)

        serialised_settings = model_settings.model_dump_json()

        self.config = ModelConfig(
            type=model_type,
            model_class=model_class.__name__,
            model_settings=serialised_settings,
            left_query=left_query.config,
            right_query=right_query.config if right_query else None,
        )

    @property
    def dependencies(self) -> list[str]:
        """Returns all resolution names this model needs as implied by the queries."""
        if self.right_query:
            return (
                self.left_query.config.dependencies
                + self.right_query.config.dependencies
            )
        return self.left_query.config.dependencies

    @property
    def parents(self) -> list[str]:
        """Returns all points of truth input to this model."""
        if self.right_query:
            return [
                self.left_query.config.point_of_truth,
                self.right_query.config.point_of_truth,
            ]
        return [self.left_query.config.point_of_truth]

    def to_resolution(self) -> Resolution:
        """Convert to Resolution for API calls."""
        return Resolution(
            name=self.name,
            description=self.description,
            truth=self._truth,
            resolution_type=ResolutionType.MODEL,
            config=self.config,
        )

    @classmethod
    def from_resolution(cls, resolution: Resolution, dag: DAG) -> "Model":
        """Reconstruct from Resolution."""
        if resolution.resolution_type != ResolutionType.MODEL:
            raise ValueError("Resolution must be of type 'model'")

        return cls(
            dag=dag,
            name=resolution.name,
            description=resolution.description,
            model_class=resolution.config.model_class,
            model_settings=resolution.config.model_settings,
            left_query=resolution.config.left_query,
            right_query=resolution.config.right_query,
            truth=resolution.truth,
        )

    @property
    def truth(self) -> float | None:
        """Returns the truth threshold for the model as a float."""
        if self._truth is not None:
            return _truth_int_to_float(self._truth)
        return None

    @truth.setter
    def truth(self, truth: float) -> None:
        """Set the truth threshold for the model."""
        self._truth = _truth_float_to_int(truth)

    def delete(self, certain: bool = False) -> bool:
        """Delete the model from the database."""
        result = _handler.delete_resolution(name=self.name, certain=certain)
        return result.success

    def run(self, for_validation: bool = False, full_rerun: bool = False) -> Results:
        """Execute the model pipeline and return results.

        Args:
            for_validation: Whether to download and store extra data to explore and
                    score results.
            full_rerun: Whether to force a re-run even if the results are cached
        """
        if self.last_run and not full_rerun:
            warnings.warn("Model already run, skipping.", UserWarning, stacklevel=2)
            return self.results

        left_df = self.left_query.run(
            return_leaf_id=for_validation,
            batch_size=settings.batch_size,
            full_rerun=full_rerun,
        )
        right_df = None

        if self.config.type == ModelType.LINKER:
            right_df = self.right_query.run(
                return_leaf_id=for_validation,
                batch_size=settings.batch_size,
                full_rerun=full_rerun,
            )

            self.model_instance.prepare(left_df, right_df)
            results = self.model_instance.link(left=left_df, right=right_df)
        else:
            self.model_instance.prepare(left_df)
            results = self.model_instance.dedupe(data=left_df)

        if for_validation:
            self.results = Results(
                probabilities=results,
                left_root_leaf=self.left_query.leaf_id.to_arrow(),
                right_root_leaf=self.right_query.leaf_id.to_arrow()
                if right_df is not None
                else None,
            )
        else:
            self.results = Results(probabilities=results)

        self.last_run = datetime.now()
        return self.results

    def sync(self) -> None:
        """Send the model config, truth and results to the server."""
        resolution = self.to_resolution()
        try:
            existing_resolution = _handler.get_resolution(name=self.name)
        except MatchboxResolutionNotFoundError:
            existing_resolution = None
        # Check if config matches
        if existing_resolution:
            if existing_resolution.config != self.config:
                raise ValueError(
                    f"Resolution {self.name} already exists with different "
                    "configuration. Please delete the existing resolution "
                    "or use a different name. "
                )
            else:
                log_prefix = f"Resolution {self.name}"
                logger.warning("Already exists. Passing.", prefix=log_prefix)
        else:
            _handler.create_resolution(resolution=resolution)

        _handler.set_truth(name=self.name, truth=self._truth)

        if self.results and len(self.results.probabilities):
            _handler.set_data(
                name=self.name,
                data=self.results.probabilities,
                validate_type=ResolutionType.MODEL,
            )

    def download_results(self) -> Results:
        """Retrieve results associated with the model from the database."""
        results = _handler.get_results(name=self.name)
        return Results(probabilities=results, metadata=self.config)

    def query(self, *sources, **kwargs) -> Query:
        """Generate a query for this model."""
        return self.dag.query(*sources, **kwargs, model=self)


def _truth_float_to_int(truth: float) -> int:
    """Convert user input float truth values to int."""
    if isinstance(truth, float) and 0.0 <= truth <= 1.0:
        return round(truth * 100)
    else:
        raise ValueError(f"Truth value {truth} not a valid probability")


def _truth_int_to_float(truth: int) -> float:
    """Convert backend int truth values to float."""
    if isinstance(truth, int) and 0 <= truth <= 100:
        return float(truth / 100)
    else:
        raise ValueError(f"Truth value {truth} not valid")



import pyarrow as pa
import pytest
from sqlalchemy import Column, MetaData, func, select
from sqlalchemy.dialects.postgresql import BIGINT, TEXT
from sqlalchemy.orm import declarative_base

from matchbox.common.exceptions import MatchboxDatabaseWriteError
from matchbox.server.postgresql import MatchboxPostgres
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.mixin import CountMixin
from matchbox.server.postgresql.orm import PKSpace
from matchbox.server.postgresql.utils.db import (
    ingest_to_temporary_table,
    large_append,
)


@pytest.mark.docker
def test_reserve_id_block(
    matchbox_postgres: MatchboxPostgres,  # Reset DB
):
    """Test that we can atomically reserve ID blocks."""
    first_cluster_id = PKSpace.reserve_block("clusters", 42)
    second_cluster_id = PKSpace.reserve_block("clusters", 42)

    assert first_cluster_id == second_cluster_id - 42

    first_cluster_keys_id = PKSpace.reserve_block("cluster_keys", 42)
    second_cluster_keys_id = PKSpace.reserve_block("cluster_keys", 42)

    assert first_cluster_keys_id == second_cluster_keys_id - 42

    with pytest.raises(ValueError):
        PKSpace.reserve_block("clusters", 0)


@pytest.mark.docker
def test_large_append(
    matchbox_postgres: MatchboxPostgres,  # will drop dummy table
):
    """Test appending large data to a table."""
    engine = MBDB.get_engine()
    metadata = MetaData(schema=MBDB.MatchboxBase.metadata.schema)

    # Initialise DummyTable to which we'll ingest
    class DummyTable(CountMixin, declarative_base(metadata=metadata)):
        __tablename__ = "dummytable"
        key = Column(BIGINT, primary_key=True)
        foo = Column(TEXT, nullable=False)

    metadata.create_all(engine, tables=[DummyTable.__table__])
    metadata.reflect(engine)
    original_tables = len(metadata.tables)

    # No auto-commit
    with MBDB.get_adbc_connection() as adbc_connection:
        large_append(
            data=pa.Table.from_pylist([{"foo": "val"}]),
            table_class=DummyTable,
            adbc_connection=adbc_connection,
        )
    assert DummyTable.count() == 0

    # Ingest data with manual keys and chunking
    with MBDB.get_adbc_connection() as adbc_connection:
        large_append(
            data=pa.Table.from_pylist(
                [{"key": 0, "foo": "val1"}],
            ),
            table_class=DummyTable,
            adbc_connection=adbc_connection,
            max_chunksize=100,
        )
        adbc_connection.commit()

    # Ingest data without keys and no chunking
    with MBDB.get_adbc_connection() as adbc_connection:
        large_append(
            data=pa.Table.from_pylist([{"foo": "val2"}]),
            table_class=DummyTable,
            adbc_connection=adbc_connection,
        )
        adbc_connection.commit()

    # Both rows were fine
    assert DummyTable.count() == 2
    with MBDB.get_session() as session:
        second_id = (
            session.query(DummyTable.key).filter(DummyTable.foo == "val2").scalar()
        )
    assert second_id > 0

    # Upserting not allowed
    with (
        pytest.raises(MatchboxDatabaseWriteError),
        MBDB.get_adbc_connection() as adbc_connection,
    ):
        large_append(
            data=pa.Table.from_pylist([{"key": 0, "foo": "val3"}]),
            table_class=DummyTable,
            adbc_connection=adbc_connection,
        )

    # Failed ingestion has no effect
    assert DummyTable.count() == 2
    assert len(metadata.tables) == original_tables

    # Columns not available in the target table are rejected
    with (
        pytest.raises(ValueError, match="does not have columns"),
        MBDB.get_adbc_connection() as adbc_connection,
    ):
        large_append(
            data=pa.Table.from_pylist([{"key": 10, "bar": "val3"}]),
            table_class=DummyTable,
            adbc_connection=adbc_connection,
        )


@pytest.mark.docker
def test_ingest_to_temporary_table(
    matchbox_postgres: MatchboxPostgres,  # will drop dummy table
):
    """Test temporary table creation, data ingestion, and automatic cleanup."""
    # Create sample arrow data
    data = pa.Table.from_pylist(
        [
            {"id": 1, "value": "test1"},
            {"id": 2, "value": "test2"},
        ]
    )

    schema_name = MBDB.MatchboxBase.metadata.schema
    table_name = "test_temp_ingest"

    # Define the column types for the temporary table
    column_types = {
        "id": BIGINT,
        "value": TEXT,
    }

    # Use the context manager to create and populate a temporary table
    with (
        ingest_to_temporary_table(
            table_name=table_name,
            schema_name=schema_name,
            data=data,
            column_types=column_types,
        ) as temp_table,
        MBDB.get_session() as session,
    ):
        # Verify the table exists and has the expected data
        # Check that the table exists using SQLAlchemy syntax
        result = session.execute(select(func.count()).select_from(temp_table)).scalar()
        assert result == 2

        # Check a specific value using SQLAlchemy syntax
        value = session.execute(
            select(temp_table.c.value).where(temp_table.c.id == 1)
        ).scalar()
        assert value == "test1"

    # After context exit, verify the table no longer exists
    with (
        MBDB.get_session() as session,
        pytest.raises(Exception),  # Should fail as table is dropped # noqa: B017
    ):
        session.execute(select(func.count()).select_from(temp_table))



from matchbox.client.models import comparison


def test_comparisons():
    comparison_name_id = comparison(
        sql_condition=(
            "l.company_name = r.company_name and l.data_hub_id = r.data_hub_id"
        )
    )

    assert comparison_name_id is not None



"""Test probabilistic behavior of linkers."""

from collections.abc import Callable
from typing import Any
from unittest.mock import Mock, patch

import numpy as np
import polars as pl
import pytest
from splink import SettingsCreator
from splink import comparison_library as cl

from matchbox.client.models import Model
from matchbox.client.models.linkers.base import Linker
from matchbox.client.models.linkers.splinklinker import SplinkLinker, SplinkSettings
from matchbox.client.models.linkers.weighteddeterministic import (
    WeightedDeterministicLinker,
    WeightedDeterministicSettings,
)
from matchbox.client.queries import Query
from matchbox.client.results import Results
from matchbox.common.factories.entities import (
    FeatureConfig,
    ReplaceRule,
    SuffixRule,
)
from matchbox.common.factories.sources import (
    SourceTestkit,
    SourceTestkitParameters,
    linked_sources_factory,
)

LinkerConfigurator = Callable[[SourceTestkit, SourceTestkit], dict[str, Any]]

# Methodology configuration adapters


def configure_weighted_probabilistic(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure WeightedDeterministicLinker with probabilistic-like behavior.

    Args:
        left_testkit: Left source object from linked_sources_factory
        right_testkit: Right source object from linked_sources_factory

    Returns:
        A dictionary with validated settings for WeightedDeterministicLinker
    """

    # Extract field names excluding key and id
    left_fields = [
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    ]
    right_fields = [
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    ]

    # Generate geometric series of weights
    weights = [1 * (0.5**i) for i in range(len(left_fields))]
    total_weight = sum(weights)

    # Normalize weights to sum to 1
    normalized_weights = [w / total_weight for w in weights]

    weighted_comparisons = []
    for l_field, r_field, weight in zip(
        left_fields, right_fields, normalized_weights, strict=False
    ):
        weighted_comparisons.append(
            {"comparison": f"l.{l_field} = r.{r_field}", "weight": weight}
        )

    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "weighted_comparisons": weighted_comparisons,
        "threshold": 0.0,
    }

    # Validate the settings dictionary
    WeightedDeterministicSettings.model_validate(settings_dict)

    return settings_dict


def configure_splink_probabilistic(
    left_testkit: SourceTestkit, right_testkit: SourceTestkit
) -> dict[str, Any]:
    """Configure SplinkLinker for probabilistic matching.

    Args:
        left_testkit: Left source object from linked_sources_factory
        right_testkit: Right source object from linked_sources_factory

    Returns:
        A dictionary with validated settings for SplinkLinker
    """

    # Extract field names excluding key and id
    left_fields = [
        c.name
        for c in left_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    ]
    right_fields = [
        c.name
        for c in right_testkit.source_config.index_fields
        if c.name not in ("key", "id")
    ]

    # Create comparison functions based on field type
    comparisons = []
    blocking_rules = []
    deterministic_matching_rules = []

    for l_field, r_field in zip(left_fields, right_fields, strict=True):
        # Splink requires exact name matches
        assert l_field == r_field

        field = l_field  # Use common field name after checking they match
        field_type = next(
            (
                c.type
                for c in left_testkit.source_config.index_fields
                if c.name == field
            ),
            "TEXT",
        )

        # Create deterministic matching rule for each field
        deterministic_matching_rules.append(f"l.{field} = r.{field}")

        # String fields
        if "TEXT" in field_type.upper():
            if field_type != "TEXT" or field[0:3].isalpha():
                blocking_rules.append(
                    f"SUBSTR(l.{field}, 1, 3) = SUBSTR(r.{field}, 1, 3)"
                )
            comparisons.append(cl.JaroWinklerAtThresholds(field, [0.9, 0.7]))

        # Numeric fields
        elif any(
            t in field_type.upper() for t in ["INT", "FLOAT", "DECIMAL", "DOUBLE"]
        ):
            blocking_rules.append(f"CAST(l.{field} AS INT) = CAST(r.{field} AS INT)")
            comparisons.append(cl.ExactMatch(field))

        else:
            comparisons.append(cl.ExactMatch(field))

    # Create Splink settings
    linker_settings = SettingsCreator(
        link_type="link_only",
        blocking_rules_to_generate_predictions=blocking_rules,
        comparisons=comparisons,
    )

    # Create training functions
    training_functions = [
        {
            "function": "estimate_probability_two_random_records_match",
            "arguments": {
                "deterministic_matching_rules": deterministic_matching_rules,
                "recall": 0.7,
            },
        },
        {
            "function": "estimate_u_using_random_sampling",
            "arguments": {"max_pairs": 1e4},
        },
    ]

    settings_dict = {
        "left_id": "id",
        "right_id": "id",
        "linker_training_functions": training_functions,
        "linker_settings": linker_settings,
        "threshold": 0.01,
    }

    # Validate the settings dictionary
    SplinkSettings.model_validate(settings_dict)

    return settings_dict


PROBABILISTIC_LINKERS = [
    pytest.param(SplinkLinker, configure_splink_probabilistic, id="Splink"),
    pytest.param(
        WeightedDeterministicLinker,
        configure_weighted_probabilistic,
        id="WeightedDeterministic",
    ),
]

# Test cases


@pytest.mark.parametrize(("Linker", "configure_linker"), PROBABILISTIC_LINKERS)
@patch.object(Query, "run")
def test_probabilistic_scores_generation(
    mock_query_run: Mock, Linker: Linker, configure_linker: LinkerConfigurator
):
    """Test that linkers can generate varying probability scores."""

    # Create sources with variations
    features = (
        FeatureConfig(
            name="company_name",
            base_generator="company",
            variations=[
                SuffixRule(suffix=" Ltd"),
                SuffixRule(suffix=" Limited"),
                ReplaceRule(old="&", new="and"),
            ],
            drop_base=False,
        ),
        FeatureConfig(
            name="id_number",
            base_generator="numerify",
            parameters=(("text", "######"),),
        ),
    )

    configs = (
        SourceTestkitParameters(
            name="source_left", features=features, n_true_entities=10
        ),
        SourceTestkitParameters(
            name="source_right", features=features, n_true_entities=10
        ),
    )

    linked = linked_sources_factory(source_parameters=configs, seed=42)
    left_source = linked.sources["source_left"]
    right_source = linked.sources["source_right"]

    mock_query_run.side_effect = [
        pl.from_arrow(left_source.data),
        pl.from_arrow(right_source.data),
    ]

    # Configure and run the linker
    linker = Model(
        dag=linked.dag,
        name="prob_test_linker",
        description="Testing probability generation",
        model_class=Linker,
        model_settings=configure_linker(left_source, right_source),
        left_query=Query(left_source, dag=linked.dag),
        right_query=Query(right_source, dag=linked.dag),
    )

    results: Results = linker.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"

    # Validate results over a threshold as a subset of the ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=left_source.entities,
        right_clusters=right_source.entities,
        sources=["source_left", "source_right"],
        threshold=np.mean(results.probabilities["probability"]),
    )

    assert not identical, f"Expected imperfect results but got: {report}"
    # Expect subsets of matches where connections weren't made
    assert report["subset"] > 0
    # Expect no wrong or invalid matches (perfect possible but unlikely)
    assert report["wrong"] == 0
    assert report["invalid"] == 0



"""DAG container for testkits."""

from pydantic import BaseModel, ConfigDict

from matchbox.client.dags import DAG
from matchbox.common.factories.models import ModelTestkit
from matchbox.common.factories.sources import LinkedSourcesTestkit, SourceTestkit
from matchbox.common.graph import (
    ModelResolutionName,
    ResolutionName,
    SourceResolutionName,
)


class TestkitDAG(BaseModel):
    """Simple DAG container for testkits."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    dag: DAG = DAG("collection")

    sources: dict[SourceResolutionName, SourceTestkit] = {}
    linked: dict[str, LinkedSourcesTestkit] = {}
    source_to_linked: dict[SourceResolutionName, str] = {}
    models: dict[ModelResolutionName, ModelTestkit] = {}

    # Dependency graph tracking
    adjacency: dict[ResolutionName, set[ResolutionName]] = {}  # -> direct dependencies
    root_source_names: dict[ResolutionName, set[SourceResolutionName]] = {}

    # Keep track of all used names to ensure uniqueness
    _all_names: set[str] = set()

    def _validate_unique_name(self, name: ResolutionName) -> str:
        """Ensure a name is unique across all testkits."""
        if name in self._all_names:
            raise ValueError(
                f"Name collision detected: '{name}' is already used in this DAG"
            )
        self._all_names.add(name)
        return name

    def _validate_dependencies(
        self, left_res: ResolutionName, right_res: ResolutionName | None
    ) -> bool:
        """Ensure dependencies are valid."""
        valid_deps = set(self.sources.keys()) | set(self.models.keys())
        dependencies = {dep for dep in [left_res, right_res] if dep is not None}

        missing_deps = dependencies - valid_deps
        if missing_deps:
            raise ValueError(
                f"Missing dependencies for model: {missing_deps}. "
                "Ensure all dependencies are added before models."
            )
        return True

    def _update_root_source_names(self, name: ModelResolutionName):
        """Update root source names for a model."""
        for dep_name in self.adjacency[name]:
            if dep_name in self.sources:
                # If dependency is a source, add it directly
                self.root_source_names[name].add(dep_name)
            elif dep_name in self.models:
                # If dependency is a model, add all its root sources
                self.root_source_names[name].update(self.root_source_names[dep_name])

    def add_source(self, testkit: SourceTestkit | LinkedSourcesTestkit):
        """Add a source testkit to the DAG."""
        if isinstance(testkit, LinkedSourcesTestkit):
            # Create linked key as "linked_" + concatenated source resolution names
            source_names = sorted(testkit.sources.keys())
            linked_key = f"linked_{'_'.join(source_names)}"

            # Store LinkedSourcesTestkit in the linked dict with the new key
            self.linked[linked_key] = testkit

            # Add each source and track its association with the linked key
            for source_name, source_testkit in testkit.sources.items():
                self._validate_unique_name(source_name)
                self.sources[source_name] = source_testkit
                self.source_to_linked[source_name] = linked_key
        else:
            source_name = self._validate_unique_name(testkit.name)
            self.sources[source_name] = testkit

    def add_model(self, testkit: ModelTestkit):
        """Add a model testkit to the DAG."""
        name = self._validate_unique_name(testkit.name)
        self.models[name] = testkit
        self.adjacency[name] = set()
        self.root_source_names[name] = set()

        # Validate dependencies
        left_res = testkit.model.config.left_query.point_of_truth
        right_res = (
            testkit.model.config.right_query.point_of_truth
            if testkit.model.config.right_query
            else None
        )
        self._validate_dependencies(left_res, right_res)

        # Track dependencies based on left/right resolution
        if left_res:
            self.adjacency[name].add(left_res)

        if right_res and right_res != left_res:
            self.adjacency[name].add(right_res)

        # Update root sources
        self._update_root_source_names(name)

    def get_sources_for_model(
        self, name: ModelResolutionName
    ) -> dict[str | None, set[SourceResolutionName]]:
        """Find the LinkedSourcesTestkit keys and specific source names for a model.

        Args:
            name: The name of the model to analyze

        Returns:
            A dictionary mapping:
            - LinkedSourcesTestkit keys (or None) to sets of model's source names
        """
        if name not in self.models:
            return {None: set()}

        root_sources = self.root_source_names.get(name, set())
        if not root_sources:
            return {None: set()}

        result = {}
        for name in root_sources:
            # linked_key could be None
            linked_key = self.source_to_linked.get(name)
            if linked_key not in result:
                result[linked_key] = set()
            result[linked_key].add(name)

        return result



import pytest

from matchbox.common.dtos import Match


def test_match_validates():
    """Match objects are validated when they're instantiated."""
    Match(
        cluster=1,
        source="test.source_config",
        source_id={"a"},
        target="test.target",
        target_id={"b"},
    )

    # Missing source_id with target_id
    with pytest.raises(ValueError):
        Match(
            cluster=1,
            source="test.source_config",
            target="test.target",
            target_id={"b"},
        )

    # Missing cluster with target_id
    with pytest.raises(ValueError):
        Match(
            source="test.source_config",
            source_id={"a"},
            target="test.target",
            target_id={"b"},
        )

    # Missing source_id with cluster
    with pytest.raises(ValueError):
        Match(
            cluster=1,
            source="test.source_config",
            target="test.target",
        )
